"""
Module for managing PowerShell through PowerShellGet (PSGet)

:depends:
    - PowerShell 5.0
    - PSGet

Support for PowerShell
"""

import logging
import xml.etree.ElementTree

import salt.utils.platform
import salt.utils.versions
from salt.exceptions import CommandExecutionError

# Set up logging
log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "psget"


def __virtual__():
    """
    Set the system module of the kernel is Windows
    """
    # Verify Windows
    if not salt.utils.platform.is_windows():
        log.debug("Module PSGet: Only available on Windows systems")
        return False, "Module PSGet: Only available on Windows systems"

    # Verify PowerShell
    powershell_info = __salt__["cmd.shell_info"]("powershell")
    if not powershell_info["installed"]:
        log.debug("Module PSGet: Requires PowerShell")
        return False, "Module PSGet: Requires PowerShell"

    # Verify PowerShell 5.0 or greater
    if salt.utils.versions.compare(powershell_info["version"], "<", "5.0"):
        log.debug("Module PSGet: Requires PowerShell 5 or newer")
        return False, "Module PSGet: Requires PowerShell 5 or newer."

    return __virtualname__


def _ps_xml_to_dict(parent, dic=None):
    """
    Formats powershell Xml to a dict.
    Note: This _ps_xml_to_dict is not perfect with powershell Xml.
    """

    if dic is None:
        dic = {}

    for child in parent:
        if list(child):
            new_dic = _ps_xml_to_dict(child, {})
            if "Name" in new_dic:
                dic[new_dic["Name"]] = new_dic
            else:
                try:
                    dic[
                        [name for ps_type, name in child.items() if ps_type == "Type"][
                            0
                        ]
                    ] = new_dic
                except IndexError:
                    dic[child.text] = new_dic
        else:
            for xml_type, name in child.items():
                if xml_type == "Name":
                    dic[name] = child.text

    return dic


def _pshell(cmd, cwd=None, depth=2):
    """
    Execute the desired powershell command and ensure that it returns data
    in Xml format and load that into python
    """

    cmd = f'{cmd} | ConvertTo-Xml -Depth {depth} -As "stream"'
    log.debug("DSC: %s", cmd)

    results = __salt__["cmd.run_all"](
        cmd, shell="powershell", cwd=cwd, python_shell=True
    )

    if "pid" in results:
        del results["pid"]

    if "retcode" not in results or results["retcode"] != 0:
        # run_all logs an error to log.error, fail hard back to the user
        raise CommandExecutionError(f"Issue executing powershell {cmd}", info=results)

    try:
        ret = _ps_xml_to_dict(
            xml.etree.ElementTree.fromstring(results["stdout"].encode("utf-8"))
        )
    except xml.etree.ElementTree.ParseError:
        results["stdout"] = results["stdout"][:1000] + ". . ."
        raise CommandExecutionError("No XML results from powershell", info=results)

    return ret


def bootstrap():
    """
    Make sure that nuget-anycpu.exe is installed.
    This will download the official nuget-anycpu.exe from the internet.

    CLI Example:

    .. code-block:: bash

        salt 'win01' psget.bootstrap
    """
    cmd = (
        "Get-PackageProvider -Name NuGet -ForceBootstrap | Select Name, Version,"
        " ProviderPath"
    )
    ret = _pshell(cmd, depth=1)
    return ret


def avail_modules(desc=False):
    """
    List available modules in registered Powershell module repositories.

    :param desc: If ``True``, the verbose description will be returned.
    :type  desc: ``bool``

    CLI Example:

    .. code-block:: bash

        salt 'win01' psget.avail_modules
        salt 'win01' psget.avail_modules desc=True
    """
    cmd = "Find-Module | Select Name, Description"
    modules = _pshell(cmd, depth=1)
    names = []
    if desc:
        names = {}
    for key in modules:
        module = modules[key]
        if desc:
            names[module["Name"]] = module["Description"]
            continue
        names.append(module["Name"])
    return names


def list_modules(desc=False):
    """
    List currently installed PSGet Modules on the system.

    :param desc: If ``True``, the verbose description will be returned.
    :type  desc: ``bool``

    CLI Example:

    .. code-block:: bash

        salt 'win01' psget.list_modules
        salt 'win01' psget.list_modules desc=True
    """
    cmd = "Get-InstalledModule"
    modules = _pshell(cmd)
    names = []
    if desc:
        names = {}
    for key in modules:
        module = modules[key]
        if desc:
            names[module["Name"]] = module
            continue
        names.append(module["Name"])
    return names


def install(
    name, minimum_version=None, required_version=None, scope=None, repository=None
):
    """
    Install a Powershell module from powershell gallery on the system.

    :param name: Name of a Powershell module
    :type  name: ``str``

    :param minimum_version: The maximum version to install, e.g. 1.23.2
    :type  minimum_version: ``str``

    :param required_version: Install a specific version
    :type  required_version: ``str``

    :param scope: The scope to install the module to, e.g. CurrentUser, Computer
    :type  scope: ``str``

    :param repository: The friendly name of a private repository, e.g. MyREpo
    :type  repository: ``str``

    CLI Example:

    .. code-block:: bash

        salt 'win01' psget.install PowerPlan
    """
    # Putting quotes around the parameter protects against command injection
    flags = [("Name", name)]

    if minimum_version is not None:
        flags.append(("MinimumVersion", minimum_version))
    if required_version is not None:
        flags.append(("RequiredVersion", required_version))
    if scope is not None:
        flags.append(("Scope", scope))
    if repository is not None:
        flags.append(("Repository", repository))
    params = ""
    for flag, value in flags:
        params += f"-{flag} {value} "
    cmd = f"Install-Module {params} -Force"
    _pshell(cmd)
    return name in list_modules()


def update(name, maximum_version=None, required_version=None):
    """
    Update a PowerShell module to a specific version, or the newest

    :param name: Name of a Powershell module
    :type  name: ``str``

    :param maximum_version: The maximum version to install, e.g. 1.23.2
    :type  maximum_version: ``str``

    :param required_version: Install a specific version
    :type  required_version: ``str``

    CLI Example:

    .. code-block:: bash

        salt 'win01' psget.update PowerPlan
    """
    # Putting quotes around the parameter protects against command injection
    flags = [("Name", name)]

    if maximum_version is not None:
        flags.append(("MaximumVersion", maximum_version))
    if required_version is not None:
        flags.append(("RequiredVersion", required_version))

    params = ""
    for flag, value in flags:
        params += f"-{flag} {value} "
    cmd = f"Update-Module {params} -Force"
    _pshell(cmd)
    return name in list_modules()


def remove(name):
    """
    Remove a Powershell DSC module from the system.

    :param  name: Name of a Powershell DSC module
    :type   name: ``str``

    CLI Example:

    .. code-block:: bash

        salt 'win01' psget.remove PowerPlan
    """
    # Putting quotes around the parameter protects against command injection
    cmd = f'Uninstall-Module "{name}"'
    no_ret = _pshell(cmd)
    return name not in list_modules()


def register_repository(name, location, installation_policy=None):
    """
    Register a PSGet repository on the local machine

    :param name: The name for the repository
    :type  name: ``str``

    :param location: The URI for the repository
    :type  location: ``str``

    :param installation_policy: The installation policy
        for packages, e.g. Trusted, Untrusted
    :type  installation_policy: ``str``

    CLI Example:

    .. code-block:: bash

        salt 'win01' psget.register_repository MyRepo https://myrepo.mycompany.com/packages
    """
    # Putting quotes around the parameter protects against command injection
    flags = [("Name", name)]

    flags.append(("SourceLocation", location))
    if installation_policy is not None:
        flags.append(("InstallationPolicy", installation_policy))

    params = ""
    for flag, value in flags:
        params += f"-{flag} {value} "
    cmd = f"Register-PSRepository {params}"
    no_ret = _pshell(cmd)
    return name not in list_modules()


def get_repository(name):
    """
    Get the details of a local PSGet repository

    :param  name: Name of the repository
    :type   name: ``str``

    CLI Example:

    .. code-block:: bash

        salt 'win01' psget.get_repository MyRepo
    """
    # Putting quotes around the parameter protects against command injection
    cmd = f'Get-PSRepository "{name}"'
    no_ret = _pshell(cmd)
    return name not in list_modules()



"""
Runner for setting and querying data via the sdb API on the master
"""

import salt.utils.sdb

__func_alias__ = {
    "set_": "set",
}


def get(uri):
    """
    Get a value from a db, using a uri in the form of sdb://<profile>/<key>. If
    the uri provided does not start with sdb://, then it will be returned as-is.

    CLI Example:

    .. code-block:: bash

        salt-run sdb.get sdb://mymemcached/foo
    """
    return salt.utils.sdb.sdb_get(uri, __opts__, __utils__)


def set_(uri, value):
    """
    Set a value in a db, using a uri in the form of ``sdb://<profile>/<key>``.
    If the uri provided does not start with ``sdb://`` or the value is not
    successfully set, return ``False``.

    CLI Example:

    .. code-block:: bash

        salt-run sdb.set sdb://mymemcached/foo bar
    """
    return salt.utils.sdb.sdb_set(uri, value, __opts__, __utils__)


def delete(uri):
    """
    Delete a value from a db, using a uri in the form of ``sdb://<profile>/<key>``.
    If the uri provided does not start with ``sdb://`` or the value is not
    successfully deleted, return ``False``.

    CLI Example:

    .. code-block:: bash

        salt-run sdb.delete sdb://mymemcached/foo
    """
    return salt.utils.sdb.sdb_delete(uri, __opts__, __utils__)


def get_or_set_hash(
    uri, length=8, chars="abcdefghijklmnopqrstuvwxyz0123456789!@#$%^&*(-_=+)"
):
    """
    Perform a one-time generation of a hash and write it to sdb.
    If that value has already been set return the value instead.

    This is useful for generating passwords or keys that are specific to
    multiple minions that need to be stored somewhere centrally.

    CLI Example:

    .. code-block:: bash

        salt-run sdb.get_or_set_hash 'SECRET_KEY' 50

    .. warning::

        This function could return strings which may contain characters which are reserved
        as directives by the YAML parser, such as strings beginning with ``%``. To avoid
        issues when using the output of this function in an SLS file containing YAML+Jinja,
        surround the call with single quotes.
    """
    return salt.utils.sdb.sdb_get_or_set_hash(uri, __opts__, length, chars, __utils__)



import logging
import os

import pytest

import salt.crypt
from tests.support.mock import MagicMock, MockCall, mock_open, patch

RSA = pytest.importorskip("Cryptodome.PublicKey.RSA")

try:
    import M2Crypto  # pylint: disable=unused-import

    HAS_M2 = True
except ImportError:
    HAS_M2 = False

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.skipif(HAS_M2, reason="m2crypto is used by salt.crypt if installed"),
]


@pytest.mark.slow_test
def test_gen_keys(tmp_path):
    key_path = str(tmp_path / "keydir")
    open_priv_wb = MockCall(os.path.join(key_path, "keyname.pem"), "wb+")
    open_pub_wb = MockCall(os.path.join(key_path, "keyname.pub"), "wb+")

    real_is_file = os.path.isfile

    def is_file(path):
        if path.startswith(str(tmp_path)):
            return False
        return real_is_file(path)

    with patch.multiple(
        os,
        umask=MagicMock(),
        chmod=MagicMock(),
        access=MagicMock(return_value=True),
    ):
        with patch("salt.utils.files.fopen", mock_open()) as m_open, patch(
            "os.path.isfile", return_value=True
        ):
            result = salt.crypt.gen_keys(key_path, "keyname", 2048)
            assert result == os.path.join(key_path, "keyname.pem")
            assert open_priv_wb not in m_open.calls
            assert open_pub_wb not in m_open.calls

        with patch("salt.utils.files.fopen", mock_open()) as m_open, patch(
            "os.path.isfile", is_file
        ):
            salt.crypt.gen_keys(key_path, "keyname", 2048)
            assert open_priv_wb in m_open.calls
            assert open_pub_wb in m_open.calls



"""
     Test cases for salt.modules.hashutil
"""

import pytest

import salt.modules.hashutil as hashutil
from tests.support.mock import patch


@pytest.fixture
def the_string():
    return "get salted"


@pytest.fixture
def the_string_base64():
    return "Z2V0IHNhbHRlZA==\n"


@pytest.fixture
def the_string_md5():
    return "2aacf29e92feaf528fb738bcf9d647ac"


@pytest.fixture
def the_string_sha256():
    return "d49859ccbc854fa68d800b5734efc70d72383e6479d545468bc300263164ff33"


@pytest.fixture
def the_string_sha512():
    return "a8c174a7941c64a068e686812a2fafd7624c840fde800f5965fbeca675f2f6e37061ffe41e17728c919bdea290eab7a21e13c04ae71661955a87f2e0e04bb045"


@pytest.fixture
def the_string_hmac():
    return "eBWf9bstXg+NiP5AOwppB5HMvZiYMPzEM9W5YMm/AmQ="


@pytest.fixture
def the_string_hmac_compute():
    return "78159ff5bb2d5e0f8d88fe403b0a690791ccbd989830fcc433d5b960c9bf0264"


@pytest.fixture
def the_string_github():
    return "sha1=b06aa56bdf4935eec82c4e53e83ed03f03fdb32d"


@pytest.fixture
def configure_loader_modules():
    return {hashutil: {}}


def test_base64_encodestring(the_string, the_string_base64):
    assert hashutil.base64_encodestring(the_string) == the_string_base64


def test_base64_decodestring(the_string, the_string_base64):
    assert hashutil.base64_decodestring(the_string_base64) == the_string


@pytest.mark.skip_on_fips_enabled_platform
def test_md5_digest(the_string, the_string_md5):
    assert hashutil.md5_digest(the_string) == the_string_md5


def test_sha256_digest(the_string, the_string_sha256):
    assert hashutil.sha256_digest(the_string) == the_string_sha256


def test_sha512_digest(the_string, the_string_sha512):
    assert hashutil.sha512_digest(the_string) == the_string_sha512


def test_hmac_signature(the_string, the_string_hmac):
    assert hashutil.hmac_signature(the_string, "shared secret", the_string_hmac)


def test_hmac_compute(the_string, the_string_hmac_compute):
    assert hashutil.hmac_compute(the_string, "shared secret")


def test_github_signature(the_string, the_string_github):
    assert hashutil.github_signature(the_string, "shared secret", the_string_github)


def test_github_signature_uses_hmac_compare_digest(the_string, the_string_github):
    with patch("hmac.compare_digest") as hmac_compare:
        assert hashutil.github_signature(the_string, "shared secret", the_string_github)
        hmac_compare.assert_called_once()



"""

    salt.serializers.tomlmod
    ~~~~~~~~~~~~~~~~~~~~~~~~

    Implements TOML serializer.

    It's just a wrapper around the python toml module.
"""

import logging

from salt.serializers import DeserializationError, SerializationError

try:
    import toml

    HAS_TOML = True
except ImportError:
    HAS_TOML = False


__virtualname__ = "toml"

__all__ = ["deserialize", "serialize", "HAS_TOML"]

log = logging.getLogger(__name__)


def __virtual__():
    if HAS_TOML is False:
        return (False, "The 'toml' library is missing")
    return __virtualname__


def deserialize(stream_or_string, **options):
    """
    Deserialize from TOML into Python data structure.

    :param stream_or_string: toml stream or string to deserialize.
    :param options: options given to the python toml module.
    """

    try:
        if not isinstance(stream_or_string, (bytes, str)):
            return toml.load(stream_or_string, **options)

        if isinstance(stream_or_string, bytes):
            stream_or_string = stream_or_string.decode("utf-8")

        return toml.loads(stream_or_string)
    except Exception as error:  # pylint: disable=broad-except
        raise DeserializationError(error)


def serialize(obj, **options):
    """
    Serialize Python data to TOML.

    :param obj: the data structure to serialize.
    :param options: options given to the python toml module.
    """

    try:
        if "file_out" in options:
            return toml.dump(obj, options["file_out"], **options)
        else:
            return toml.dumps(obj, **options)
    except Exception as error:  # pylint: disable=broad-except
        raise SerializationError(error)



import shutil
import ssl
import tarfile

import pytest
from pytestshellutils.utils import ports
from saltfactories.utils import random_string

import salt.utils.http

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("docker", "dockerd", check_all=False),
]


@pytest.mark.parametrize("backend", ["requests", "urllib2", "tornado"])
def test_decode_body(webserver, integration_files_dir, backend):
    with tarfile.open(integration_files_dir / "test.tar.gz", "w:gz") as tar:
        tar.add(integration_files_dir / "this.txt")

    ret = salt.utils.http.query(
        webserver.url("test.tar.gz"), backend=backend, decode_body=False
    )
    assert isinstance(ret["body"], bytes)


@pytest.fixture(scope="module")
def tinyproxy_port():
    return ports.get_unused_localhost_port()


@pytest.fixture(scope="module")
def tinyproxy_user():
    return random_string("tinyproxy-user-")


@pytest.fixture(scope="module")
def tinyproxy_pass():
    return random_string("tinyproxy-pass-")


@pytest.fixture(scope="session")
def ca():
    trustme = pytest.importorskip("trustme")
    return trustme.CA()


@pytest.fixture(scope="session")
def httpserver_ssl_context(ca):
    context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
    localhost_cert = ca.issue_cert("127.0.0.1")
    localhost_cert.configure_cert(context)
    return context


@pytest.fixture(scope="session")
def httpclient_ssl_context(ca):
    with ca.cert_pem.tempfile() as ca_temp_path:
        return ssl.create_default_context(cafile=ca_temp_path)


@pytest.fixture(params=[True, False], ids=lambda x: "basic-auth" if x else "no-auth")
def tinyproxy_basic_auth(request):
    return request.param


@pytest.fixture(params=[True, False], ids=lambda x: "no-proxy" if x else "with-proxy")
def no_proxy(request):
    return request.param


@pytest.fixture(params=["POST", "GET"], ids=lambda x: x)
def http_method(request):
    return request.param


@pytest.fixture(scope="module")
def tinyproxy_dir(tmp_path_factory):
    try:
        dirname = tmp_path_factory.mktemp("tinyproxy")
        yield dirname
    finally:
        shutil.rmtree(dirname, ignore_errors=True)


@pytest.fixture
def tinyproxy_conf(
    tinyproxy_dir, tinyproxy_port, tinyproxy_user, tinyproxy_pass, tinyproxy_basic_auth
):
    basic_auth = (
        f"\nBasicAuth {tinyproxy_user} {tinyproxy_pass}" if tinyproxy_basic_auth else ""
    )
    conf = """Port {port}
Listen 127.0.0.1
Timeout 600
Allow 127.0.0.1
AddHeader "X-Tinyproxy-Header" "Test custom tinyproxy header"{auth}
    """.format(
        port=tinyproxy_port, auth=basic_auth
    )
    (tinyproxy_dir / "tinyproxy.conf").write_text(conf)


@pytest.fixture
def tinyproxy_container(
    salt_factories,
    tinyproxy_conf,
    tinyproxy_dir,
):
    container = salt_factories.get_container(
        "tinyproxy",
        image_name="ghcr.io/saltstack/salt-ci-containers/tinyproxy:latest",
        container_run_kwargs={
            "network_mode": "host",
            "volumes": {str(tinyproxy_dir): {"bind": "/etc/tinyproxy", "mode": "z"}},
        },
        pull_before_start=True,
        skip_on_pull_failure=True,
        skip_if_docker_client_not_connectable=True,
    )
    with container.started() as factory:
        yield factory


@pytest.mark.parametrize("backend", ["requests", "tornado", "urllib2"])
def test_real_proxy(
    tinyproxy_container,
    httpserver,
    ca,
    tinyproxy_port,
    tinyproxy_user,
    tinyproxy_pass,
    backend,
    tinyproxy_basic_auth,
    no_proxy,
    http_method,
):
    data = b"mydatahere"
    opts = {
        "proxy_host": "localhost",
        "proxy_port": tinyproxy_port,
    }
    if tinyproxy_basic_auth:
        opts.update(
            {
                "proxy_username": tinyproxy_user,
                "proxy_password": tinyproxy_pass,
            }
        )

    # Expecting the headers allows verification that it went through the proxy without looking at the logs
    if no_proxy:
        opts["no_proxy"] = ["random.hostname.io", httpserver.host]
        httpserver.expect_request(
            "/real_proxy_test",
        ).respond_with_data(data, content_type="application/octet-stream")
    else:
        httpserver.expect_request(
            "/real_proxy_test",
        ).respond_with_data(data, content_type="application/octet-stream")
    url = httpserver.url_for("/real_proxy_test").replace("localhost", "127.0.0.1")

    # We just want to be sure that it's using the proxy
    with ca.cert_pem.tempfile() as ca_temp_path:
        ret = salt.utils.http.query(
            url,
            method=http_method,
            data=data,
            backend=backend,
            opts=opts,
            decode_body=False,
            verify_ssl=ca_temp_path,
        )
    body = ret.get("body", "")
    assert body == data



"""
cache Module

:maintainer:    SaltStack
:maturity:      New
:platform:      all

.. versionadded:: 2017.7.0

This module provides access to Salt's cache subsystem.

Like all sdb modules, the cache module requires a configuration profile to
be configured in either the minion or master configuration file. This profile
requires very little. In the example:

.. code-block:: yaml

    mastercloudcache:
      driver: cache
      bank: cloud/active/ec2/my-ec2-conf/saltmaster
      cachedir: /var/cache/salt

The ``driver`` refers to the cache module, ``bank`` refers to the cache bank
that contains the data and ``cachedir`` (optional), if used, points to an
alternate directory for cache data storage.

.. code-block:: yaml

    master_ip: sdb://mastercloudcache/public_ips

It is also possible to override both the ``bank`` and ``cachedir`` options
inside the SDB URI:

.. code-block:: yaml

    master_ip: sdb://mastercloudcache/public_ips?cachedir=/var/cache/salt

For this reason, both the ``bank`` and the ``cachedir`` options can be
omitted from the SDB profile. However, if the ``bank`` option is omitted,
it must be specified in the URI:

.. code-block:: yaml

    master_ip: sdb://mastercloudcache/public_ips?bank=cloud/active/ec2/my-ec2-conf/saltmaster
"""

import salt.cache

__func_alias__ = {"set_": "set"}

__virtualname__ = "cache"


def __virtual__():
    """
    Only load the module if keyring is installed
    """
    return __virtualname__


def set_(key, value, service=None, profile=None):  # pylint: disable=W0613
    """
    Set a key/value pair in the cache service
    """
    key, profile = _parse_key(key, profile)
    cache = salt.cache.Cache(__opts__)
    cache.store(profile["bank"], key, value)
    return get(key, service, profile)


def get(key, service=None, profile=None):  # pylint: disable=W0613
    """
    Get a value from the cache service
    """
    key, profile = _parse_key(key, profile)
    cache = salt.cache.Cache(__opts__)
    return cache.fetch(profile["bank"], key=key)


def delete(key, service=None, profile=None):  # pylint: disable=W0613
    """
    Get a value from the cache service
    """
    key, profile = _parse_key(key, profile)
    cache = salt.cache.Cache(__opts__)
    try:
        cache.flush(profile["bank"], key=key)
        return True
    except Exception:  # pylint: disable=broad-except
        return False


def _parse_key(key, profile):
    """
    Parse out a key and update the opts with any override data
    """
    comps = key.split("?")
    if len(comps) > 1:
        for item in comps[1].split("&"):
            newkey, newval = item.split("=")
            profile[newkey] = newval
    if "cachedir" in profile:
        __opts__["cachedir"] = profile["cachedir"]
    return comps[0], profile



import shutil

import pytest

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
]


@pytest.fixture(scope="module", autouse=True)
def pillar_tree(base_env_pillar_tree_root_dir):
    top_file = """
    base:
      'localhost':
        - mine
      '127.0.0.1':
        - mine
    """
    mine_pillar_file = """
    mine_functions:
      disk.usage:
        - c
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_pillar_tree_root_dir
    )
    mine_tempfile = pytest.helpers.temp_file(
        "mine.sls", mine_pillar_file, base_env_pillar_tree_root_dir
    )

    with top_tempfile, mine_tempfile:
        yield


@pytest.fixture(autouse=True)
def thin_dir(salt_ssh_cli):
    try:
        yield
    finally:
        ret = salt_ssh_cli.run("config.get", "thin_dir")
        assert ret.returncode == 0
        thin_dir_path = ret.data
        shutil.rmtree(thin_dir_path, ignore_errors=True)


def test_ssh_mine_get(salt_ssh_cli):
    """
    test salt-ssh with mine
    """
    ret = salt_ssh_cli.run("mine.get", "localhost", "test.arg")
    assert ret.returncode == 0
    assert ret.data
    assert "localhost" in ret.data
    assert "args" in ret.data["localhost"]
    assert ret.data["localhost"]["args"] == ["itworked"]


@pytest.mark.parametrize("tgts", (("ssh",), ("regular",), ("ssh", "regular")))
def test_mine_get(salt_ssh_cli, salt_minion, tgts):
    """
    Test mine returns with both regular and SSH minions
    """
    if len(tgts) > 1:
        tgt = "*"
        exp = {"localhost", salt_minion.id}
    else:
        tgt = "localhost" if "ssh" in tgts else salt_minion.id
        exp = {tgt}
    ret = salt_ssh_cli.run(
        "mine.get",
        "*",
        "test.ping",
        ssh_minions="ssh" in tgts,
        regular_minions="regular" in tgts,
    )
    assert ret.returncode == 0
    assert ret.data
    assert set(ret.data) == exp
    for id_ in exp:
        assert ret.data[id_] is True


def test_ssh_mine_get_error(salt_ssh_cli, caplog):
    """
    Test that a mine function returning an error is not
    included in the output.
    """
    ret = salt_ssh_cli.run("mine.get", "localhost", "disk.usage")
    assert ret.returncode == 0
    assert not ret.data
    assert "Error executing mine func disk.usage" in caplog.text



import logging
import subprocess
import time

import pytest

log = logging.getLogger(__name__)


@pytest.mark.skip_unless_on_linux
def test_dns_change(master, minion, salt_cli, etc_hosts, caplog, master_alive_interval):
    """
    Verify a minion will pick up a master's dns change if it's been disconnected.
    """

    etc_hosts.write_text(f"{etc_hosts.orig_text}\n172.16.0.1    master.local")

    with minion.started(start_timeout=180):
        with caplog.at_level(logging.INFO):
            ret = salt_cli.run("test.ping", minion_tgt="minion")
            assert ret.returncode == 0
            etc_hosts.write_text(f"{etc_hosts.orig_text}\n127.0.0.1    master.local")
            log.info("Changed hosts record for master1.local and master2.local")
            subprocess.check_output(["ip", "addr", "del", "172.16.0.1/32", "dev", "lo"])
            log.info("Removed secondary master IP address.")
            # Wait for the minion's master_alive_interval, adding a second for
            # reliablity.
            time.sleep(master_alive_interval + 1)
            assert (
                "Master ip address changed from 172.16.0.1 to 127.0.0.1" in caplog.text
            )
            ret = salt_cli.run("test.ping", minion_tgt="minion")
            assert ret.returncode == 0



"""
Test AnsibleGate State Module
"""

import logging
import shutil
import textwrap

import pytest
import yaml
from pytestshellutils.exceptions import FactoryTimeout
from saltfactories.utils.functional import StateResult

import salt.utils.files
import salt.utils.path
from tests.support.runtests import RUNTIME_VARS

pytestmark = [
    # root access is necessary because the playbooks install/uninstall software
    pytest.mark.skip_if_not_root,
    # Because of the above, these are also destructive tests
    pytest.mark.destructive_test,
    pytest.mark.skip_if_binaries_missing(
        "ansible-playbook", reason="ansible-playbook is not installed"
    ),
]

log = logging.getLogger(__name__)


@pytest.fixture(scope="module")
def ansible_inventory_directory(tmp_path_factory, grains):
    if grains["os_family"] != "RedHat" or grains["os"] == "VMware Photon OS":
        pytest.skip("Currently, the test targets the RedHat OS familly only.")
    tmp_dir = tmp_path_factory.mktemp("ansible")
    try:
        yield tmp_dir
    finally:
        shutil.rmtree(str(tmp_dir))


@pytest.fixture(scope="module", autouse=True)
def ansible_inventory(ansible_inventory_directory, sshd_server, known_hosts_file):
    inventory = str(ansible_inventory_directory / "inventory")
    client_key = str(sshd_server.config_dir / "client_key")
    data = {
        "all": {
            "hosts": {
                "localhost": {
                    "ansible_host": "127.0.0.1",
                    "ansible_port": sshd_server.listen_port,
                    "ansible_user": RUNTIME_VARS.RUNNING_TESTS_USER,
                    "ansible_ssh_private_key_file": client_key,
                    "ansible_ssh_extra_args": (
                        f"-o UserKnownHostsFile={known_hosts_file} "
                    ),
                },
            },
        },
    }
    with salt.utils.files.fopen(inventory, "w") as yaml_file:
        yaml.dump(data, yaml_file, default_flow_style=False)
    return inventory


@pytest.mark.requires_sshd_server
@pytest.mark.timeout_unless_on_windows(240)
def test_ansible_playbook(salt_call_cli, ansible_inventory, tmp_path):
    rundir = tmp_path / "rundir"
    rundir.mkdir(exist_ok=True, parents=True)
    remove_contents = textwrap.dedent(
        """
    ---
    - hosts: all
      tasks:
      - name: remove postfix
        yum:
          name: postfix
          state: absent
        become: true
        become_user: root
    """
    )
    remove_playbook = rundir / "remove.yml"
    remove_playbook.write_text(remove_contents)
    install_contents = textwrap.dedent(
        """
    ---
    - hosts: all
      tasks:
      - name: install postfix
        yum:
          name: postfix
          state: present
        become: true
        become_user: root
    """
    )
    install_playbook = rundir / "install.yml"
    install_playbook.write_text(install_contents)

    # These tests have been known to timeout, so allow them longer if needed.
    timeouts = [60, 120, 180]
    names = ["remove.yml", "install.yml"]

    for name in names:
        for timeout in timeouts:
            try:
                ret = salt_call_cli.run(
                    "state.single",
                    "ansible.playbooks",
                    name=name,
                    rundir=str(rundir),
                    ansible_kwargs={"inventory": ansible_inventory},
                    _timeout=timeout,  # The removal can take over 60 seconds
                )
            except FactoryTimeout:
                log.debug("%s took longer than %s seconds", name, timeout)
                if timeout == timeouts[-1]:
                    pytest.fail(f"Failed to run {name}")
            else:
                assert ret.returncode == 0
                assert StateResult(ret.data).result is True
                break



"""
    :codeauthor: :email:`Amel Ajdinovic <amel@stackpointcloud.com>`
"""

import pytest

from tests.integration.cloud.helpers.cloud_test_base import TIMEOUT, CloudTest

pytest.importorskip("oneandone.client", reason="salt-cloud requires >= 1and1 1.2.0")


class OneAndOneTest(CloudTest):
    """
    Integration tests for the 1and1 cloud provider
    """

    PROVIDER = "oneandone"
    REQUIRED_PROVIDER_CONFIG_ITEMS = ("api_token",)

    def test_list_images(self):
        """
        Tests the return of running the --list-images command for 1and1
        """
        image_list = self.run_cloud(f"--list-images {self.PROVIDER_NAME}")
        self.assertIn("coreOSimage", [i.strip() for i in image_list])

    def test_instance(self):
        """
        Test creating an instance on 1and1
        """
        # check if instance with salt installed returned
        ret_str = self.run_cloud(
            f"-p oneandone-test {self.instance_name}", timeout=TIMEOUT
        )
        self.assertInstanceExists(ret_str)

        self.assertDestroyInstance()



"""
Module for managing disks and blockdevices
"""

import collections
import decimal
import logging
import os
import re
import subprocess

import salt.utils.decorators
import salt.utils.decorators.path
import salt.utils.path
import salt.utils.platform
from salt.exceptions import CommandExecutionError

__func_alias__ = {"format_": "format"}

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only work on POSIX-like systems
    """
    if salt.utils.platform.is_windows():
        return False, "This module doesn't work on Windows."
    return True


def _parse_numbers(text):
    """
    Convert a string to a number, allowing for a K|M|G|T postfix, 32.8K.
    Returns a decimal number if the string is a real number,
    or the string unchanged otherwise.
    """
    if text.isdigit():
        return decimal.Decimal(text)

    try:
        postPrefixes = {
            "K": "10E3",
            "M": "10E6",
            "G": "10E9",
            "T": "10E12",
            "P": "10E15",
            "E": "10E18",
            "Z": "10E21",
            "Y": "10E24",
        }
        if text[-1] in postPrefixes:
            v = decimal.Decimal(text[:-1])
            v = v * decimal.Decimal(postPrefixes[text[-1]])
            return v
        else:
            return decimal.Decimal(text)
    except ValueError:
        return text


def _clean_flags(args, caller):
    """
    Sanitize flags passed into df
    """
    flags = ""
    if args is None:
        return flags
    # TODO: most of these cause the result parsing to fail
    allowed = ("a", "B", "h", "H", "i", "k", "l", "P", "t", "T", "x", "v")
    for flag in args:
        if flag in allowed:
            flags += flag
        else:
            raise CommandExecutionError(f"Invalid flag passed to {caller}")
    return flags


def usage(args=None):
    """
    Return usage information for volumes mounted on this minion

    args
        Sequence of flags to pass to the ``df`` command.

    .. versionchanged:: 2019.2.0

        Default for SunOS changed to 1 kilobyte blocks

    CLI Example:

    .. code-block:: bash

        salt '*' disk.usage
    """
    flags = _clean_flags(args, "disk.usage")
    if not os.path.isfile("/etc/mtab") and __grains__["kernel"] == "Linux":
        log.error("df cannot run without /etc/mtab")
        if __grains__.get("virtual_subtype") == "LXC":
            log.error(
                "df command failed and LXC detected. If you are running "
                "a Docker container, consider linking /proc/mounts to "
                "/etc/mtab or consider running Docker with -privileged"
            )
        return {}
    if __grains__["kernel"] == "Linux":
        cmd = "df -P"
    elif __grains__["kernel"] == "OpenBSD" or __grains__["kernel"] == "AIX":
        cmd = "df -kP"
    elif __grains__["kernel"] == "SunOS":
        cmd = "df -k"
    else:
        cmd = "df"
    if flags:
        cmd += f" -{flags}"
    ret = {}
    out = __salt__["cmd.run"](cmd, python_shell=False).splitlines()
    oldline = None
    for line in out:
        if not line:
            continue
        if line.startswith("Filesystem"):
            continue
        if oldline:
            line = oldline + " " + line
        comps = line.split()
        if len(comps) == 1:
            oldline = line
            continue
        else:
            oldline = None
        while len(comps) >= 2 and not comps[1].isdigit():
            comps[0] = f"{comps[0]} {comps[1]}"
            comps.pop(1)
        if len(comps) < 2:
            continue
        try:
            if __grains__["kernel"] == "Darwin":
                ret[comps[8]] = {
                    "filesystem": comps[0],
                    "512-blocks": comps[1],
                    "used": comps[2],
                    "available": comps[3],
                    "capacity": comps[4],
                    "iused": comps[5],
                    "ifree": comps[6],
                    "%iused": comps[7],
                }
            else:
                ret[comps[5]] = {
                    "filesystem": comps[0],
                    "1K-blocks": comps[1],
                    "used": comps[2],
                    "available": comps[3],
                    "capacity": comps[4],
                }
        except IndexError:
            log.error("Problem parsing disk usage information")
            ret = {}
    return ret


def inodeusage(args=None):
    """
    Return inode usage information for volumes mounted on this minion

    args
        Sequence of flags to pass to the ``df`` command.

    CLI Example:

    .. code-block:: bash

        salt '*' disk.inodeusage
    """
    flags = _clean_flags(args, "disk.inodeusage")
    if __grains__["kernel"] == "AIX":
        cmd = "df -i"
    else:
        cmd = "df -iP"
    if flags:
        cmd += f" -{flags}"
    ret = {}
    out = __salt__["cmd.run"](cmd, python_shell=False).splitlines()
    for line in out:
        if line.startswith("Filesystem"):
            continue
        comps = line.split()
        # Don't choke on empty lines
        if not comps:
            continue

        try:
            if __grains__["kernel"] == "OpenBSD":
                ret[comps[8]] = {
                    "inodes": int(comps[5]) + int(comps[6]),
                    "used": comps[5],
                    "free": comps[6],
                    "use": comps[7],
                    "filesystem": comps[0],
                }
            elif __grains__["kernel"] == "AIX":
                ret[comps[6]] = {
                    "inodes": comps[4],
                    "used": comps[5],
                    "free": comps[2],
                    "use": comps[5],
                    "filesystem": comps[0],
                }
            else:
                ret[comps[5]] = {
                    "inodes": comps[1],
                    "used": comps[2],
                    "free": comps[3],
                    "use": comps[4],
                    "filesystem": comps[0],
                }
        except (IndexError, ValueError):
            log.error("Problem parsing inode usage information")
            ret = {}
    return ret


def percent(args=None):
    """
    Return partition information for volumes mounted on this minion

    args
        Specify a single partition for which to return data.

    CLI Example:

    .. code-block:: bash

        salt '*' disk.percent /var
    """
    if __grains__["kernel"] == "Linux":
        cmd = "df -P"
    elif __grains__["kernel"] == "OpenBSD" or __grains__["kernel"] == "AIX":
        cmd = "df -kP"
    else:
        cmd = "df"
    ret = {}
    out = __salt__["cmd.run"](cmd, python_shell=False).splitlines()
    for line in out:
        if not line:
            continue
        if line.startswith("Filesystem"):
            continue
        comps = line.split()
        while len(comps) >= 2 and not comps[1].isdigit():
            comps[0] = f"{comps[0]} {comps[1]}"
            comps.pop(1)
        if len(comps) < 2:
            continue
        try:
            if __grains__["kernel"] == "Darwin":
                ret[comps[8]] = comps[4]
            else:
                ret[comps[5]] = comps[4]
        except IndexError:
            log.error("Problem parsing disk usage information")
            ret = {}
    if args and args not in ret:
        log.error(
            "Problem parsing disk usage information: Partition '%s' does not exist!",
            args,
        )
        ret = {}
    elif args:
        return ret[args]

    return ret


@salt.utils.decorators.path.which("blkid")
def blkid(device=None, token=None):
    """
    Return block device attributes: UUID, LABEL, etc. This function only works
    on systems where blkid is available.

    device
        Device name from the system

    token
        Any valid token used for the search

    CLI Example:

    .. code-block:: bash

        salt '*' disk.blkid
        salt '*' disk.blkid /dev/sda
        salt '*' disk.blkid token='UUID=6a38ee5-7235-44e7-8b22-816a403bad5d'
        salt '*' disk.blkid token='TYPE=ext4'
    """
    cmd = ["blkid"]
    if device:
        cmd.append(device)
    elif token:
        cmd.extend(["-t", token])

    ret = {}
    blkid_result = __salt__["cmd.run_all"](cmd, python_shell=False)

    if blkid_result["retcode"] > 0:
        return ret

    for line in blkid_result["stdout"].splitlines():
        if not line:
            continue
        comps = line.split()
        device = comps[0][:-1]
        info = {}
        device_attributes = re.split('"*"', line.partition(" ")[2])
        for key, value in zip(*[iter(device_attributes)] * 2):
            key = key.strip("=").strip(" ")
            info[key] = value.strip('"')
        ret[device] = info

    return ret


def tune(device, **kwargs):
    """
    Set attributes for the specified device

    CLI Example:

    .. code-block:: bash

        salt '*' disk.tune /dev/sda1 read-ahead=1024 read-write=True

    Valid options are: ``read-ahead``, ``filesystem-read-ahead``,
    ``read-only``, ``read-write``.

    See the ``blockdev(8)`` manpage for a more complete description of these
    options.
    """

    kwarg_map = {
        "read-ahead": "setra",
        "filesystem-read-ahead": "setfra",
        "read-only": "setro",
        "read-write": "setrw",
    }
    opts = ""
    args = []
    for key in kwargs:
        if key in kwarg_map:
            switch = kwarg_map[key]
            if key != "read-write":
                args.append(switch.replace("set", "get"))
            else:
                args.append("getro")
            if kwargs[key] == "True" or kwargs[key] is True:
                opts += f"--{key} "
            else:
                opts += f"--{switch} {kwargs[key]} "
    cmd = f"blockdev {opts}{device}"
    out = __salt__["cmd.run"](cmd, python_shell=False).splitlines()
    return dump(device, args)


def wipe(device):
    """
    Remove the filesystem information

    CLI Example:

    .. code-block:: bash

        salt '*' disk.wipe /dev/sda1
    """

    cmd = f"wipefs -a {device}"
    try:
        out = __salt__["cmd.run_all"](cmd, python_shell=False)
    except subprocess.CalledProcessError as err:
        return False
    if out["retcode"] == 0:
        return True
    else:
        log.error("Error wiping device %s: %s", device, out["stderr"])
        return False


def dump(device, args=None):
    """
    Return all contents of dumpe2fs for a specified device

    device
        The device path to dump.

    args
        A list of attributes to return. Returns all by default.

    CLI Example:

    .. code-block:: bash

        salt '*' disk.dump /dev/sda1
    """
    cmd = (
        "blockdev --getro --getsz --getss --getpbsz --getiomin --getioopt --getalignoff"
        " --getmaxsect --getsize --getsize64 --getra --getfra {}".format(device)
    )
    ret = {}
    opts = [c[2:] for c in cmd.split() if c.startswith("--")]
    out = __salt__["cmd.run_all"](cmd, python_shell=False)
    if out["retcode"] == 0:
        lines = [line for line in out["stdout"].splitlines() if line]
        count = 0
        for line in lines:
            ret[opts[count]] = line
            count = count + 1
        if args:
            temp_ret = {}
            for arg in args:
                temp_ret[arg] = ret[arg]
            return temp_ret
        else:
            return ret
    else:
        return False


def resize2fs(device):
    """
    Resizes the filesystem.

    CLI Example:

    .. code-block:: bash

        salt '*' disk.resize2fs /dev/sda1
    """
    cmd = f"resize2fs {device}"
    try:
        out = __salt__["cmd.run_all"](cmd, python_shell=False)
    except subprocess.CalledProcessError as err:
        return False
    if out["retcode"] == 0:
        return True


@salt.utils.decorators.path.which("sync")
@salt.utils.decorators.path.which("mkfs")
def format_(
    device,
    fs_type="ext4",
    inode_size=None,
    lazy_itable_init=None,
    fat=None,
    force=False,
):
    """
    Format a filesystem onto a device

    .. versionadded:: 2016.11.0

    device
        The device in which to create the new filesystem

    fs_type
        The type of filesystem to create

    inode_size
        Size of the inodes

        This option is only enabled for ext and xfs filesystems

    lazy_itable_init
        If enabled and the uninit_bg feature is enabled, the inode table will
        not be fully initialized by mke2fs.  This speeds up filesystem
        initialization noticeably, but it requires the kernel to finish
        initializing the filesystem  in  the  background  when  the filesystem
        is first mounted.  If the option value is omitted, it defaults to 1 to
        enable lazy inode table zeroing.

        This option is only enabled for ext filesystems

    fat
        FAT size option. Can be 12, 16 or 32, and can only be used on
        fat or vfat filesystems.

    force
        Force mke2fs to create a filesystem, even if the specified device is
        not a partition on a block special device. This option is only enabled
        for ext and xfs filesystems

        This option is dangerous, use it with caution.

    CLI Example:

    .. code-block:: bash

        salt '*' disk.format /dev/sdX1
    """
    cmd = ["mkfs", "-t", str(fs_type)]
    if inode_size is not None:
        if fs_type[:3] == "ext":
            cmd.extend(["-i", str(inode_size)])
        elif fs_type == "xfs":
            cmd.extend(["-i", f"size={inode_size}"])
    if lazy_itable_init is not None:
        if fs_type[:3] == "ext":
            cmd.extend(["-E", f"lazy_itable_init={lazy_itable_init}"])
    if fat is not None and fat in (12, 16, 32):
        if fs_type[-3:] == "fat":
            cmd.extend(["-F", fat])
    if force:
        if fs_type[:3] == "ext":
            cmd.append("-F")
        elif fs_type == "xfs":
            cmd.append("-f")
    cmd.append(str(device))

    mkfs_success = __salt__["cmd.retcode"](cmd, ignore_retcode=True) == 0
    sync_success = __salt__["cmd.retcode"]("sync", ignore_retcode=True) == 0

    return all([mkfs_success, sync_success])


@salt.utils.decorators.path.which_bin(["lsblk", "df"])
def fstype(device):
    """
    Return the filesystem name of the specified device

    .. versionadded:: 2016.11.0

    device
        The name of the device

    CLI Example:

    .. code-block:: bash

        salt '*' disk.fstype /dev/sdX1
    """
    if salt.utils.path.which("lsblk"):
        lsblk_out = __salt__["cmd.run"](f"lsblk -o fstype {device}").splitlines()
        if len(lsblk_out) > 1:
            fs_type = lsblk_out[1].strip()
            if fs_type:
                return fs_type

    if salt.utils.path.which("df"):
        # the fstype was not set on the block device, so inspect the filesystem
        # itself for its type
        if __grains__["kernel"] == "AIX" and os.path.isfile("/usr/sysv/bin/df"):
            df_out = __salt__["cmd.run"](f"/usr/sysv/bin/df -n {device}").split()
            if len(df_out) > 2:
                fs_type = df_out[2]
                if fs_type:
                    return fs_type
        else:
            df_out = __salt__["cmd.run"](f"df -T {device}").splitlines()
            if len(df_out) > 1:
                fs_type = df_out[1]
                if fs_type:
                    return fs_type

    return ""


@salt.utils.decorators.path.which("hdparm")
def _hdparm(args, failhard=True):
    """
    Execute hdparm
    Fail hard when required
    return output when possible
    """
    cmd = f"hdparm {args}"
    result = __salt__["cmd.run_all"](cmd)
    if result["retcode"] != 0:
        msg = "{}: {}".format(cmd, result["stderr"])
        if failhard:
            raise CommandExecutionError(msg)
        else:
            log.warning(msg)

    return result["stdout"]


@salt.utils.decorators.path.which("hdparm")
def hdparms(disks, args="aAbBcCdgHiJkMmNnQrRuW"):
    """
    Retrieve disk parameters.

    .. versionadded:: 2016.3.0

    disks
        Single disk or list of disks to query.

    args
        Sequence of ``hdparm`` flags to fetch.

    CLI Example:

    .. code-block:: bash

        salt '*' disk.hdparms /dev/sda
    """
    if isinstance(args, (list, tuple)):
        args = "".join(args)

    if not isinstance(disks, (list, tuple)):
        disks = [disks]

    out = {}
    for disk in disks:
        if not disk.startswith("/dev"):
            disk = f"/dev/{disk}"
        disk_data = {}
        for line in _hdparm(f"-{args} {disk}", False).splitlines():
            line = line.strip()
            if not line or line == disk + ":":
                continue

            if ":" in line:
                key, vals = line.split(":", 1)
                key = re.sub(r" is$", "", key)
            elif "=" in line:
                key, vals = line.split("=", 1)
            else:
                continue
            key = key.strip().lower().replace(" ", "_")
            vals = vals.strip()

            rvals = []
            if re.match(r"[0-9]+ \(.*\)", vals):
                vals = vals.split(" ")
                rvals.append(int(vals[0]))
                rvals.append(vals[1].strip("()"))
            else:
                valdict = {}
                for val in re.split(r"[/,]", vals.strip()):
                    val = val.strip()
                    try:
                        val = int(val)
                        rvals.append(val)
                    except Exception:  # pylint: disable=broad-except
                        if "=" in val:
                            deep_key, val = val.split("=", 1)
                            deep_key = deep_key.strip()
                            val = val.strip()
                            if val:
                                valdict[deep_key] = val
                        elif val:
                            rvals.append(val)
                if valdict:
                    rvals.append(valdict)
                if not rvals:
                    continue
                elif len(rvals) == 1:
                    rvals = rvals[0]
            disk_data[key] = rvals

        out[disk] = disk_data

    return out


@salt.utils.decorators.path.which("hdparm")
def hpa(disks, size=None):
    """
    Get/set Host Protected Area settings

    T13 INCITS 346-2001 (1367D) defines the BEER (Boot Engineering Extension Record)
    and PARTIES (Protected Area Run Time Interface Extension Services), allowing
    for a Host Protected Area on a disk.

    It's often used by OEMS to hide parts of a disk, and for overprovisioning SSD's

    .. warning::
        Setting the HPA might clobber your data, be very careful with this on active disks!

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' disk.hpa /dev/sda
        salt '*' disk.hpa /dev/sda 5%
        salt '*' disk.hpa /dev/sda 10543256
    """

    hpa_data = {}
    for disk, data in hdparms(disks, "N").items():
        visible, total, status = next(iter(data.values()))
        if visible == total or "disabled" in status:
            hpa_data[disk] = {"total": total}
        else:
            hpa_data[disk] = {
                "total": total,
                "visible": visible,
                "hidden": total - visible,
            }

    if size is None:
        return hpa_data

    for disk, data in hpa_data.items():
        try:
            size = data["total"] - int(size)
        except Exception:  # pylint: disable=broad-except
            if "%" in size:
                size = int(size.strip("%"))
                size = (100 - size) * data["total"]
                size /= 100
        if size <= 0:
            size = data["total"]

        _hdparm(f"--yes-i-know-what-i-am-doing -Np{size} {disk}")


def smart_attributes(dev, attributes=None, values=None):
    """
    Fetch SMART attributes
    Providing attributes will deliver only requested attributes
    Providing values will deliver only requested values for attributes

    Default is the Backblaze recommended
    set (https://www.backblaze.com/blog/hard-drive-smart-stats/):
    (5,187,188,197,198)

    .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' disk.smart_attributes /dev/sda
        salt '*' disk.smart_attributes /dev/sda attributes=(5,187,188,197,198)
    """

    if not dev.startswith("/dev/"):
        dev = "/dev/" + dev

    cmd = f"smartctl --attributes {dev}"
    smart_result = __salt__["cmd.run_all"](cmd, output_loglevel="quiet")
    if smart_result["retcode"] != 0:
        raise CommandExecutionError(smart_result["stderr"])

    smart_result = iter(smart_result["stdout"].splitlines())

    fields = []
    for line in smart_result:
        if line.startswith("ID#"):
            fields = re.split(r"\s+", line.strip())
            fields = [key.lower() for key in fields[1:]]
            break

    if values is not None:
        fields = [field if field in values else "_" for field in fields]

    smart_attr = {}
    for line in smart_result:
        if not re.match(r"[\s]*\d", line):
            break

        line = re.split(r"\s+", line.strip(), maxsplit=len(fields))
        attr = int(line[0])

        if attributes is not None and attr not in attributes:
            continue

        data = dict(zip(fields, line[1:]))
        try:
            del data["_"]
        except Exception:  # pylint: disable=broad-except
            pass

        for field in data:
            val = data[field]
            try:
                val = int(val)
            except Exception:  # pylint: disable=broad-except
                try:
                    val = [int(value) for value in val.split(" ")]
                except Exception:  # pylint: disable=broad-except
                    pass
            data[field] = val

        smart_attr[attr] = data

    return smart_attr


@salt.utils.decorators.path.which("iostat")
def iostat(interval=1, count=5, disks=None):
    """
    Gather and return (averaged) IO stats.

    .. versionadded:: 2016.3.0

    .. versionchanged:: 2016.11.4
        Added support for AIX

    CLI Example:

    .. code-block:: bash

        salt '*' disk.iostat 1 5 disks=sda
    """
    if salt.utils.platform.is_linux():
        return _iostat_linux(interval, count, disks)
    elif salt.utils.platform.is_freebsd():
        return _iostat_fbsd(interval, count, disks)
    elif salt.utils.platform.is_aix():
        return _iostat_aix(interval, count, disks)


def _iostats_dict(header, stats):
    """
    Transpose collected data, average it, stomp it in dict using header

    Use Decimals so we can properly calc & round, convert to float 'caus' we can't transmit Decimals over 0mq
    """
    stats = [
        float((sum(stat) / len(stat)).quantize(decimal.Decimal(".01")))
        for stat in zip(*stats)
    ]
    stats = dict(zip(header, stats))
    return stats


def _iostat_fbsd(interval, count, disks):
    """
    Tested on FreeBSD, quite likely other BSD's only need small changes in cmd syntax
    """
    if disks is None:
        iostat_cmd = f"iostat -xC -w {interval} -c {count} "
    elif isinstance(disks, str):
        iostat_cmd = f"iostat -x -w {interval} -c {count} {disks}"
    else:
        iostat_cmd = "iostat -x -w {} -c {} {}".format(interval, count, " ".join(disks))

    sys_stats = []
    dev_stats = collections.defaultdict(list)
    sys_header = []
    dev_header = []
    h_len = 1000  # randomly absurdly high

    ret = iter(
        __salt__["cmd.run_stdout"](iostat_cmd, output_loglevel="quiet").splitlines()
    )
    for line in ret:
        if not line.startswith("device"):
            continue
        elif not dev_header:
            dev_header = line.split()[1:]
        while line is not False:
            line = next(ret, False)
            if not line or not line[0].isalnum():
                break
            line = line.split()
            disk = line[0]
            stats = [decimal.Decimal(x) for x in line[1:]]
            # h_len will become smallest number of fields in stat lines
            if len(stats) < h_len:
                h_len = len(stats)
            dev_stats[disk].append(stats)

    iostats = {}

    # The header was longer than the smallest number of fields
    # Therefore the sys stats are hidden in there
    if h_len < len(dev_header):
        sys_header = dev_header[h_len:]
        dev_header = dev_header[0:h_len]

        for disk, stats in dev_stats.items():
            if len(stats[0]) > h_len:
                sys_stats = [stat[h_len:] for stat in stats]
                dev_stats[disk] = [stat[0:h_len] for stat in stats]

        iostats["sys"] = _iostats_dict(sys_header, sys_stats)

    for disk, stats in dev_stats.items():
        iostats[disk] = _iostats_dict(dev_header, stats)

    return iostats


def _iostat_linux(interval, count, disks):
    if disks is None:
        iostat_cmd = f"iostat -x {interval} {count} "
    elif isinstance(disks, str):
        iostat_cmd = f"iostat -xd {interval} {count} {disks}"
    else:
        iostat_cmd = "iostat -xd {} {} {}".format(interval, count, " ".join(disks))

    sys_stats = []
    dev_stats = collections.defaultdict(list)
    sys_header = []
    dev_header = []

    ret = iter(
        __salt__["cmd.run_stdout"](iostat_cmd, output_loglevel="quiet").splitlines()
    )
    for line in ret:
        if line.startswith("avg-cpu:"):
            if not sys_header:
                sys_header = tuple(line.split()[1:])
            line = [decimal.Decimal(x) for x in next(ret).split()]
            sys_stats.append(line)
        elif line.startswith("Device:"):
            if not dev_header:
                dev_header = tuple(line.split()[1:])
            while line is not False:
                line = next(ret, False)
                if not line or not line[0].isalnum():
                    break
                line = line.split()
                disk = line[0]
                stats = [decimal.Decimal(x) for x in line[1:]]
                dev_stats[disk].append(stats)

    iostats = {}

    if sys_header:
        iostats["sys"] = _iostats_dict(sys_header, sys_stats)

    for disk, stats in dev_stats.items():
        iostats[disk] = _iostats_dict(dev_header, stats)

    return iostats


def _iostat_aix(interval, count, disks):
    """
    AIX support to gather and return (averaged) IO stats.
    """
    log.debug("AIX disk iostat entry")

    if disks is None:
        iostat_cmd = f"iostat -dD {interval} {count} "
    elif isinstance(disks, str):
        iostat_cmd = f"iostat -dD {disks} {interval} {count}"
    else:
        iostat_cmd = "iostat -dD {} {} {}".format(" ".join(disks), interval, count)

    ret = {}
    procn = None
    fields = []
    disk_name = ""
    disk_mode = ""
    dev_stats = collections.defaultdict(list)
    for line in __salt__["cmd.run"](iostat_cmd).splitlines():
        # Note: iostat -dD is per-system
        #
        # root@l490vp031_pub:~/devtest# iostat -dD hdisk6 1 3
        #
        # System configuration: lcpu=8 drives=1 paths=2 vdisks=2
        #
        # hdisk6          xfer:  %tm_act      bps      tps      bread      bwrtn
        #                          0.0      0.0      0.0        0.0        0.0
        #                read:      rps  avgserv  minserv  maxserv   timeouts      fails
        #                          0.0      0.0      0.0      0.0           0          0
        #               write:      wps  avgserv  minserv  maxserv   timeouts      fails
        #                          0.0      0.0      0.0      0.0           0          0
        #               queue:  avgtime  mintime  maxtime  avgwqsz    avgsqsz     sqfull
        #                          0.0      0.0      0.0      0.0        0.0         0.0
        # --------------------------------------------------------------------------------
        #
        # hdisk6          xfer:  %tm_act      bps      tps      bread      bwrtn
        #                          9.6     16.4K     4.0       16.4K       0.0
        #                read:      rps  avgserv  minserv  maxserv   timeouts      fails
        #                          4.0      4.9      0.3      9.9           0          0
        #               write:      wps  avgserv  minserv  maxserv   timeouts      fails
        #                          0.0      0.0      0.0      0.0           0          0
        #               queue:  avgtime  mintime  maxtime  avgwqsz    avgsqsz     sqfull
        #                          0.0      0.0      0.0      0.0        0.0         0.0
        # --------------------------------------------------------------------------------
        #
        # hdisk6          xfer:  %tm_act      bps      tps      bread      bwrtn
        #                          0.0      0.0      0.0        0.0        0.0
        #                read:      rps  avgserv  minserv  maxserv   timeouts      fails
        #                          0.0      0.0      0.3      9.9           0          0
        #               write:      wps  avgserv  minserv  maxserv   timeouts      fails
        #                          0.0      0.0      0.0      0.0           0          0
        #               queue:  avgtime  mintime  maxtime  avgwqsz    avgsqsz     sqfull
        #                          0.0      0.0      0.0      0.0        0.0         0.0
        # --------------------------------------------------------------------------------
        if not line or line.startswith("System") or line.startswith("-----------"):
            continue

        if not re.match(r"\s", line):
            # seen disk name
            dsk_comps = line.split(":")
            dsk_firsts = dsk_comps[0].split()
            disk_name = dsk_firsts[0]
            disk_mode = dsk_firsts[1]
            fields = dsk_comps[1].split()
            if disk_name not in dev_stats.keys():
                dev_stats[disk_name] = []
                procn = len(dev_stats[disk_name])
                dev_stats[disk_name].append({})
                dev_stats[disk_name][procn][disk_mode] = {}
                dev_stats[disk_name][procn][disk_mode]["fields"] = fields
                dev_stats[disk_name][procn][disk_mode]["stats"] = []
            continue

        if ":" in line:
            comps = line.split(":")
            fields = comps[1].split()
            disk_mode = comps[0].lstrip()
            if disk_mode not in dev_stats[disk_name][0].keys():
                dev_stats[disk_name][0][disk_mode] = {}
                dev_stats[disk_name][0][disk_mode]["fields"] = fields
                dev_stats[disk_name][0][disk_mode]["stats"] = []
        else:
            line = line.split()
            stats = [_parse_numbers(x) for x in line[:]]
            dev_stats[disk_name][0][disk_mode]["stats"].append(stats)

    iostats = {}

    for disk, list_modes in dev_stats.items():
        iostats[disk] = {}
        for modes in list_modes:
            for disk_mode in modes.keys():
                fields = modes[disk_mode]["fields"]
                stats = modes[disk_mode]["stats"]
                iostats[disk][disk_mode] = _iostats_dict(fields, stats)

    return iostats


def get_fstype_from_path(path):
    """
    Return the filesystem type of the underlying device for a specified path.

    .. versionadded:: 3006.0

    path
        The path for the function to evaluate.

    CLI Example:

    .. code-block:: bash

        salt '*' disk.get_fstype_from_path /root
    """
    dev = __salt__["mount.get_device_from_path"](path)
    return fstype(dev)



import pytest

from salt.cloud.clouds import qingcloud
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        qingcloud: {
            "__opts__": {
                "providers": {
                    "qingcloud": {
                        "qingcloud": {
                            "access_key_id": "key_1234",
                            "secret_access_key": "1234",
                            "zone": "test_zone",
                            "key_filename": "/testfilename",
                            "driver": "qingcloud",
                        }
                    }
                },
                "profiles": {"qingcloud": {}},
            },
            "__active_provider_name__": "qingcloud:qingcloud",
        },
    }


def test_qingcloud_verify_ssl():
    """
    test qinglcoud when using verify_ssl
    """
    patch_sig = patch("salt.cloud.clouds.qingcloud._compute_signature", MagicMock())

    for verify in [True, False, None]:
        mock_requests = MagicMock()
        mock_requests.return_value.status_code = 200
        mock_requests.return_value.text = '{"ret_code": 0}'
        patch_requests = patch("requests.get", mock_requests)
        with patch.dict(
            qingcloud.__opts__["providers"]["qingcloud"]["qingcloud"],
            {"verify_ssl": verify},
        ):
            with patch_sig, patch_requests:
                ret = qingcloud.query()
                assert ret["ret_code"] == 0
                assert mock_requests.call_args_list[0].kwargs["verify"] == verify



"""
salt.utils.clean_path works as expected
"""

import salt.utils.verify
from tests.support.mock import patch


def test_clean_path_valid(tmp_path):
    path_a = str(tmp_path / "foo")
    path_b = str(tmp_path / "foo" / "bar")
    assert salt.utils.verify.clean_path(path_a, path_b) == path_b


def test_clean_path_invalid(tmp_path):
    path_a = str(tmp_path / "foo")
    path_b = str(tmp_path / "baz" / "bar")
    assert salt.utils.verify.clean_path(path_a, path_b) == ""


def test_clean_path_relative_root(tmp_path):
    with patch("os.getcwd", return_value=str(tmp_path)):
        path_a = "foo"
        path_b = str(tmp_path / "foo" / "bar")
        assert salt.utils.verify.clean_path(path_a, path_b) == path_b



"""
Grains for detecting what type of package Salt is using

.. versionadded:: 3007.0
"""

import logging

import salt.utils.package

log = logging.getLogger(__name__)


__virtualname__ = "package"


def __virtual__():
    return __virtualname__


def package():
    """
    Function to determine if the user is currently using
    onedir, pip or system level package of Salt.
    """
    return {"package": salt.utils.package.pkg_type()}



"""
The Saltutil module is used to manage the state of the salt minion itself. It
is used to manage minion modules as well as automate updates to the salt
minion.

:depends:   - esky Python module for update functionality
"""

import copy
import fnmatch
import logging
import os
import shutil
import signal
import sys
import time
import urllib.error

import salt
import salt.channel.client
import salt.client
import salt.client.ssh.client
import salt.config
import salt.defaults.events
import salt.payload
import salt.runner
import salt.state
import salt.utils.args
import salt.utils.event
import salt.utils.extmods
import salt.utils.files
import salt.utils.functools
import salt.utils.minion
import salt.utils.path
import salt.utils.process
import salt.utils.url
import salt.wheel
from salt.exceptions import (
    CommandExecutionError,
    SaltInvocationError,
    SaltRenderError,
    SaltReqTimeoutError,
)

try:
    import esky
    from esky import EskyVersionError

    HAS_ESKY = True
except ImportError:
    HAS_ESKY = False

try:
    import psutil

    HAS_PSUTIL = True
except ImportError:
    HAS_PSUTIL = False

# Fix a nasty bug with Win32 Python not supporting all of the standard signals
try:
    salt_SIGKILL = signal.SIGKILL
except AttributeError:
    salt_SIGKILL = signal.SIGTERM


__proxyenabled__ = ["*"]

log = logging.getLogger(__name__)

TOP_ENVS_CKEY = "saltutil._top_file_envs"


def _get_top_file_envs():
    """
    Get all environments from the top file
    """
    try:
        return __context__[TOP_ENVS_CKEY]
    except KeyError:
        with salt.state.HighState(__opts__, initial_pillar=__pillar__.value()) as st_:
            try:
                top = st_.get_top()
                if top:
                    envs = list(st_.top_matches(top).keys()) or "base"
                else:
                    envs = "base"
            except SaltRenderError as exc:
                raise CommandExecutionError(f"Unable to render top file(s): {exc}")
        __context__[TOP_ENVS_CKEY] = envs
        return envs


def _sync(form, saltenv=None, extmod_whitelist=None, extmod_blacklist=None):
    """
    Sync the given directory in the given environment
    """
    if saltenv is None:
        saltenv = _get_top_file_envs()
    if isinstance(saltenv, str):
        saltenv = saltenv.split(",")
    ret, touched = salt.utils.extmods.sync(
        __opts__,
        form,
        saltenv=saltenv,
        extmod_whitelist=extmod_whitelist,
        extmod_blacklist=extmod_blacklist,
    )
    # Dest mod_dir is touched? trigger reload if requested
    if touched:
        mod_file = os.path.join(__opts__["cachedir"], "module_refresh")
        with salt.utils.files.fopen(mod_file, "a"):
            pass
    if (
        form == "grains"
        and __opts__.get("grains_cache")
        and os.path.isfile(os.path.join(__opts__["cachedir"], "grains.cache.p"))
    ):
        try:
            os.remove(os.path.join(__opts__["cachedir"], "grains.cache.p"))
        except OSError:
            log.error("Could not remove grains cache!")
    return ret


def update(version=None):
    """
    Update the salt minion from the URL defined in opts['update_url']
    Broadcom, Inc provides the latest builds here:
    update_url: https://packages.broadcom.com/artifactory/saltproject-generic/windows/

    Be aware that as of 2014-8-11 there's a bug in esky such that only the
    latest version available in the update_url can be downloaded and installed.

    This feature requires the minion to be running a bdist_esky build.

    The version number is optional and will default to the most recent version
    available at opts['update_url'].

    Returns details about the transaction upon completion.

    CLI Examples:

    .. code-block:: bash

        salt '*' saltutil.update
        salt '*' saltutil.update 0.10.3
    """
    ret = {}
    if not HAS_ESKY:
        ret["_error"] = "Esky not available as import"
        return ret
    if not getattr(sys, "frozen", False):
        ret["_error"] = "Minion is not running an Esky build"
        return ret
    if not __salt__["config.option"]("update_url"):
        ret["_error"] = '"update_url" not configured on this minion'
        return ret
    app = esky.Esky(sys.executable, __opts__["update_url"])
    oldversion = __grains__["saltversion"]
    if not version:
        try:
            version = app.find_update()
        except urllib.error.URLError as exc:
            ret["_error"] = f"Could not connect to update_url. Error: {exc}"
            return ret
    if not version:
        ret["_error"] = "No updates available"
        return ret
    try:
        app.fetch_version(version)
    except EskyVersionError as exc:
        ret["_error"] = f"Unable to fetch version {version}. Error: {exc}"
        return ret
    try:
        app.install_version(version)
    except EskyVersionError as exc:
        ret["_error"] = f"Unable to install version {version}. Error: {exc}"
        return ret
    try:
        app.cleanup()
    except Exception as exc:  # pylint: disable=broad-except
        ret["_error"] = f"Unable to cleanup. Error: {exc}"
    restarted = {}
    for service in __opts__["update_restart_services"]:
        restarted[service] = __salt__["service.restart"](service)
    ret["comment"] = f"Updated from {oldversion} to {version}"
    ret["restarted"] = restarted
    return ret


def sync_beacons(
    saltenv=None, refresh=True, extmod_whitelist=None, extmod_blacklist=None
):
    """
    .. versionadded:: 2015.5.1

    Sync beacons from ``salt://_beacons`` to the minion

    saltenv
        The fileserver environment from which to sync. To sync from more than
        one environment, pass a comma-separated list.

        If not passed, then all environments configured in the :ref:`top files
        <states-top>` will be checked for beacons to sync. If no top files are
        found, then the ``base`` environment will be synced.

    refresh : True
        If ``True``, refresh the available beacons on the minion. This refresh
        will be performed even if no new beacons are synced. Set to ``False``
        to prevent this refresh.

    extmod_whitelist : None
        comma-separated list of modules to sync

    extmod_blacklist : None
        comma-separated list of modules to blacklist based on type

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.sync_beacons
        salt '*' saltutil.sync_beacons saltenv=dev
        salt '*' saltutil.sync_beacons saltenv=base,dev
    """
    ret = _sync("beacons", saltenv, extmod_whitelist, extmod_blacklist)
    if refresh:
        refresh_beacons()
    return ret


def sync_sdb(saltenv=None, extmod_whitelist=None, extmod_blacklist=None):
    """
    .. versionadded:: 2015.5.8,2015.8.3

    Sync sdb modules from ``salt://_sdb`` to the minion

    saltenv
        The fileserver environment from which to sync. To sync from more than
        one environment, pass a comma-separated list.

        If not passed, then all environments configured in the :ref:`top files
        <states-top>` will be checked for sdb modules to sync. If no top files
        are found, then the ``base`` environment will be synced.

    extmod_whitelist : None
        comma-separated list of modules to sync

    extmod_blacklist : None
        comma-separated list of modules to blacklist based on type

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.sync_sdb
        salt '*' saltutil.sync_sdb saltenv=dev
        salt '*' saltutil.sync_sdb saltenv=base,dev
    """
    ret = _sync("sdb", saltenv, extmod_whitelist, extmod_blacklist)
    return ret


def sync_modules(
    saltenv=None, refresh=True, extmod_whitelist=None, extmod_blacklist=None
):
    """
    .. versionadded:: 0.10.0

    Sync execution modules from ``salt://_modules`` to the minion

    saltenv
        The fileserver environment from which to sync. To sync from more than
        one environment, pass a comma-separated list.

        If not passed, then all environments configured in the :ref:`top files
        <states-top>` will be checked for execution modules to sync. If no top
        files are found, then the ``base`` environment will be synced.

    refresh : True
        If ``True``, refresh the available execution modules on the minion.
        This refresh will be performed even if no new execution modules are
        synced. Set to ``False`` to prevent this refresh.

    .. important::

        If this function is executed using a :py:func:`module.run
        <salt.states.module.run>` state, the SLS file will not have access to
        newly synced execution modules unless a ``refresh`` argument is
        added to the state, like so:

        .. code-block:: yaml

            load_my_custom_module:
              module.run:
                - name: saltutil.sync_modules
                - refresh: True

        See :ref:`here <reloading-modules>` for a more detailed explanation of
        why this is necessary.

    extmod_whitelist : None
        comma-separated list of modules to sync

    extmod_blacklist : None
        comma-separated list of modules to blacklist based on type

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.sync_modules
        salt '*' saltutil.sync_modules saltenv=dev
        salt '*' saltutil.sync_modules saltenv=base,dev
    """
    ret = _sync("modules", saltenv, extmod_whitelist, extmod_blacklist)
    if refresh:
        refresh_modules()
    return ret


def sync_states(
    saltenv=None, refresh=True, extmod_whitelist=None, extmod_blacklist=None
):
    """
    .. versionadded:: 0.10.0

    Sync state modules from ``salt://_states`` to the minion

    saltenv
        The fileserver environment from which to sync. To sync from more than
        one environment, pass a comma-separated list.

        If not passed, then all environments configured in the :ref:`top files
        <states-top>` will be checked for state modules to sync. If no top
        files are found, then the ``base`` environment will be synced.

    refresh : True
        If ``True``, refresh the available states on the minion. This refresh
        will be performed even if no new state modules are synced. Set to
        ``False`` to prevent this refresh.

    extmod_whitelist : None
        comma-separated list of modules to sync

    extmod_blacklist : None
        comma-separated list of modules to blacklist based on type

    CLI Examples:

    .. code-block:: bash

        salt '*' saltutil.sync_states
        salt '*' saltutil.sync_states saltenv=dev
        salt '*' saltutil.sync_states saltenv=base,dev
    """
    ret = _sync("states", saltenv, extmod_whitelist, extmod_blacklist)
    if refresh:
        refresh_modules()
    return ret


def refresh_grains(**kwargs):
    """
    .. versionadded:: 2016.3.6,2016.11.4,2017.7.0

    Refresh the minion's grains without syncing custom grains modules from
    ``salt://_grains``.

    .. note::
        The available execution modules will be reloaded as part of this
        proceess, as grains can affect which modules are available.

    refresh_pillar : True
        Set to ``False`` to keep pillar data from being refreshed.

    clean_pillar_cache : False
        Set to ``True`` to refresh pillar cache.

    CLI Examples:

    .. code-block:: bash

        salt '*' saltutil.refresh_grains
    """
    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    _refresh_pillar = kwargs.pop("refresh_pillar", True)
    clean_pillar_cache = kwargs.pop("clean_pillar_cache", False)
    if kwargs:
        salt.utils.args.invalid_kwargs(kwargs)
    # Modules and pillar need to be refreshed in case grains changes affected
    # them, and the module refresh process reloads the grains and assigns the
    # newly-reloaded grains to each execution module's __grains__ dunder.
    if _refresh_pillar:
        # we don't need to call refresh_modules here because it's done by refresh_pillar
        refresh_pillar(clean_cache=clean_pillar_cache)
    else:
        refresh_modules()
    return True


def sync_grains(
    saltenv=None,
    refresh=True,
    extmod_whitelist=None,
    extmod_blacklist=None,
    clean_pillar_cache=False,
):
    """
    .. versionadded:: 0.10.0

    Sync grains modules from ``salt://_grains`` to the minion

    saltenv
        The fileserver environment from which to sync. To sync from more than
        one environment, pass a comma-separated list.

        If not passed, then all environments configured in the :ref:`top files
        <states-top>` will be checked for grains modules to sync. If no top
        files are found, then the ``base`` environment will be synced.

    refresh : True
        If ``True``, refresh the available execution modules and recompile
        pillar data for the minion. This refresh will be performed even if no
        new grains modules are synced. Set to ``False`` to prevent this
        refresh.

    extmod_whitelist : None
        comma-separated list of modules to sync

    extmod_blacklist : None
        comma-separated list of modules to blacklist based on type

    clean_pillar_cache : False
        Set to ``True`` to refresh pillar cache.

    CLI Examples:

    .. code-block:: bash

        salt '*' saltutil.sync_grains
        salt '*' saltutil.sync_grains saltenv=dev
        salt '*' saltutil.sync_grains saltenv=base,dev
    """
    ret = _sync("grains", saltenv, extmod_whitelist, extmod_blacklist)
    if refresh:
        # we don't need to call refresh_modules here because it's done by refresh_pillar
        refresh_pillar(clean_cache=clean_pillar_cache)
    return ret


def sync_renderers(
    saltenv=None, refresh=True, extmod_whitelist=None, extmod_blacklist=None
):
    """
    .. versionadded:: 0.10.0

    Sync renderers from ``salt://_renderers`` to the minion

    saltenv
        The fileserver environment from which to sync. To sync from more than
        one environment, pass a comma-separated list.

        If not passed, then all environments configured in the :ref:`top files
        <states-top>` will be checked for renderers to sync. If no top files
        are found, then the ``base`` environment will be synced.

    refresh : True
        If ``True``, refresh the available execution modules on the minion.
        This refresh will be performed even if no new renderers are synced.
        Set to ``False`` to prevent this refresh.

    extmod_whitelist : None
        comma-separated list of modules to sync

    extmod_blacklist : None
        comma-separated list of modules to blacklist based on type

    CLI Examples:

    .. code-block:: bash

        salt '*' saltutil.sync_renderers
        salt '*' saltutil.sync_renderers saltenv=dev
        salt '*' saltutil.sync_renderers saltenv=base,dev
    """
    ret = _sync("renderers", saltenv, extmod_whitelist, extmod_blacklist)
    if refresh:
        refresh_modules()
    return ret


def sync_returners(
    saltenv=None, refresh=True, extmod_whitelist=None, extmod_blacklist=None
):
    """
    .. versionadded:: 0.10.0

    Sync returners from ``salt://_returners`` to the minion

    saltenv
        The fileserver environment from which to sync. To sync from more than
        one environment, pass a comma-separated list.

        If not passed, then all environments configured in the :ref:`top files
        <states-top>` will be checked for returners to sync. If no top files
        are found, then the ``base`` environment will be synced.

    refresh : True
        If ``True``, refresh the available execution modules on the minion.
        This refresh will be performed even if no new returners are synced. Set
        to ``False`` to prevent this refresh.

    extmod_whitelist : None
        comma-separated list of modules to sync

    extmod_blacklist : None
        comma-separated list of modules to blacklist based on type

    CLI Examples:

    .. code-block:: bash

        salt '*' saltutil.sync_returners
        salt '*' saltutil.sync_returners saltenv=dev
    """
    ret = _sync("returners", saltenv, extmod_whitelist, extmod_blacklist)
    if refresh:
        refresh_modules()
    return ret


def sync_proxymodules(
    saltenv=None, refresh=False, extmod_whitelist=None, extmod_blacklist=None
):
    """
    .. versionadded:: 2015.8.2

    Sync proxy modules from ``salt://_proxy`` to the minion

    saltenv
        The fileserver environment from which to sync. To sync from more than
        one environment, pass a comma-separated list.

        If not passed, then all environments configured in the :ref:`top files
        <states-top>` will be checked for proxy modules to sync. If no top
        files are found, then the ``base`` environment will be synced.

    refresh : True
        If ``True``, refresh the available execution modules on the minion.
        This refresh will be performed even if no new proxy modules are synced.
        Set to ``False`` to prevent this refresh.

    extmod_whitelist : None
        comma-separated list of modules to sync

    extmod_blacklist : None
        comma-separated list of modules to blacklist based on type

    CLI Examples:

    .. code-block:: bash

        salt '*' saltutil.sync_proxymodules
        salt '*' saltutil.sync_proxymodules saltenv=dev
        salt '*' saltutil.sync_proxymodules saltenv=base,dev
    """
    ret = _sync("proxy", saltenv, extmod_whitelist, extmod_blacklist)
    if refresh:
        refresh_modules()
    return ret


def sync_matchers(
    saltenv=None, refresh=False, extmod_whitelist=None, extmod_blacklist=None
):
    """
    .. versionadded:: 2019.2.0

    Sync engine modules from ``salt://_matchers`` to the minion

    saltenv
        The fileserver environment from which to sync. To sync from more than
        one environment, pass a comma-separated list.

        If not passed, then all environments configured in the :ref:`top files
        <states-top>` will be checked for engines to sync. If no top files are
        found, then the ``base`` environment will be synced.

    refresh : True
        If ``True``, refresh the available execution modules on the minion.
        This refresh will be performed even if no new matcher modules are synced.
        Set to ``False`` to prevent this refresh.

    extmod_whitelist : None
        comma-separated list of modules to sync

    extmod_blacklist : None
        comma-separated list of modules to blacklist based on type

    CLI Examples:

    .. code-block:: bash

        salt '*' saltutil.sync_matchers
        salt '*' saltutil.sync_matchers saltenv=base,dev
    """
    ret = _sync("matchers", saltenv, extmod_whitelist, extmod_blacklist)
    if refresh:
        refresh_modules()
    return ret


def sync_engines(
    saltenv=None, refresh=False, extmod_whitelist=None, extmod_blacklist=None
):
    """
    .. versionadded:: 2016.3.0

    Sync engine modules from ``salt://_engines`` to the minion

    saltenv
        The fileserver environment from which to sync. To sync from more than
        one environment, pass a comma-separated list.

        If not passed, then all environments configured in the :ref:`top files
        <states-top>` will be checked for engines to sync. If no top files are
        found, then the ``base`` environment will be synced.

    refresh : True
        If ``True``, refresh the available execution modules on the minion.
        This refresh will be performed even if no new engine modules are synced.
        Set to ``False`` to prevent this refresh.

    extmod_whitelist : None
        comma-separated list of modules to sync

    extmod_blacklist : None
        comma-separated list of modules to blacklist based on type

    CLI Examples:

    .. code-block:: bash

        salt '*' saltutil.sync_engines
        salt '*' saltutil.sync_engines saltenv=base,dev
    """
    ret = _sync("engines", saltenv, extmod_whitelist, extmod_blacklist)
    if refresh:
        refresh_modules()
    return ret


def sync_thorium(
    saltenv=None, refresh=False, extmod_whitelist=None, extmod_blacklist=None
):
    """
    .. versionadded:: 2018.3.0

    Sync Thorium modules from ``salt://_thorium`` to the minion

    saltenv
        The fileserver environment from which to sync. To sync from more than
        one environment, pass a comma-separated list.

        If not passed, then all environments configured in the :ref:`top files
        <states-top>` will be checked for engines to sync. If no top files are
        found, then the ``base`` environment will be synced.

    refresh: ``True``
        If ``True``, refresh the available execution modules on the minion.
        This refresh will be performed even if no new Thorium modules are synced.
        Set to ``False`` to prevent this refresh.

    extmod_whitelist
        comma-separated list of modules to sync

    extmod_blacklist
        comma-separated list of modules to blacklist based on type

    CLI Examples:

    .. code-block:: bash

        salt '*' saltutil.sync_thorium
        salt '*' saltutil.sync_thorium saltenv=base,dev
    """
    ret = _sync("thorium", saltenv, extmod_whitelist, extmod_blacklist)
    if refresh:
        refresh_modules()
    return ret


def sync_output(
    saltenv=None, refresh=True, extmod_whitelist=None, extmod_blacklist=None
):
    """
    Sync outputters from ``salt://_output`` to the minion

    saltenv
        The fileserver environment from which to sync. To sync from more than
        one environment, pass a comma-separated list.

        If not passed, then all environments configured in the :ref:`top files
        <states-top>` will be checked for outputters to sync. If no top files
        are found, then the ``base`` environment will be synced.

    refresh : True
        If ``True``, refresh the available execution modules on the minion.
        This refresh will be performed even if no new outputters are synced.
        Set to ``False`` to prevent this refresh.

    extmod_whitelist : None
        comma-separated list of modules to sync

    extmod_blacklist : None
        comma-separated list of modules to blacklist based on type

    CLI Examples:

    .. code-block:: bash

        salt '*' saltutil.sync_output
        salt '*' saltutil.sync_output saltenv=dev
        salt '*' saltutil.sync_output saltenv=base,dev
    """
    ret = _sync("output", saltenv, extmod_whitelist, extmod_blacklist)
    if refresh:
        refresh_modules()
    return ret


sync_outputters = salt.utils.functools.alias_function(sync_output, "sync_outputters")


def sync_clouds(
    saltenv=None, refresh=True, extmod_whitelist=None, extmod_blacklist=None
):
    """
    .. versionadded:: 2017.7.0

    Sync cloud modules from ``salt://_cloud`` to the minion

    saltenv : base
        The fileserver environment from which to sync. To sync from more than
        one environment, pass a comma-separated list.

    refresh : True
        If ``True``, refresh the available execution modules on the minion.
        This refresh will be performed even if no new utility modules are
        synced. Set to ``False`` to prevent this refresh.

    extmod_whitelist : None
        comma-separated list of modules to sync

    extmod_blacklist : None
        comma-separated list of modules to blacklist based on type

    CLI Examples:

    .. code-block:: bash

        salt '*' saltutil.sync_clouds
        salt '*' saltutil.sync_clouds saltenv=dev
        salt '*' saltutil.sync_clouds saltenv=base,dev
    """
    ret = _sync("clouds", saltenv, extmod_whitelist, extmod_blacklist)
    if refresh:
        refresh_modules()
    return ret


def sync_utils(
    saltenv=None, refresh=True, extmod_whitelist=None, extmod_blacklist=None
):
    """
    .. versionadded:: 2014.7.0

    Sync utility modules from ``salt://_utils`` to the minion

    saltenv
        The fileserver environment from which to sync. To sync from more than
        one environment, pass a comma-separated list.

        If not passed, then all environments configured in the :ref:`top files
        <states-top>` will be checked for utility modules to sync. If no top
        files are found, then the ``base`` environment will be synced.

    refresh : True
        If ``True``, refresh the available execution modules on the minion.
        This refresh will be performed even if no new utility modules are
        synced. Set to ``False`` to prevent this refresh.

    extmod_whitelist : None
        comma-separated list of modules to sync

    extmod_blacklist : None
        comma-separated list of modules to blacklist based on type

    CLI Examples:

    .. code-block:: bash

        salt '*' saltutil.sync_utils
        salt '*' saltutil.sync_utils saltenv=dev
        salt '*' saltutil.sync_utils saltenv=base,dev
    """
    ret = _sync("utils", saltenv, extmod_whitelist, extmod_blacklist)
    if refresh:
        refresh_modules()
    return ret


def sync_serializers(
    saltenv=None, refresh=True, extmod_whitelist=None, extmod_blacklist=None
):
    """
    .. versionadded:: 2019.2.0

    Sync serializers from ``salt://_serializers`` to the minion

    saltenv
        The fileserver environment from which to sync. To sync from more than
        one environment, pass a comma-separated list.

        If not passed, then all environments configured in the :ref:`top files
        <states-top>` will be checked for serializer modules to sync. If no top
        files are found, then the ``base`` environment will be synced.

    refresh : True
        If ``True``, refresh the available execution modules on the minion.
        This refresh will be performed even if no new serializer modules are
        synced. Set to ``False`` to prevent this refresh.

    extmod_whitelist : None
        comma-seperated list of modules to sync

    extmod_blacklist : None
        comma-seperated list of modules to blacklist based on type

    CLI Examples:

    .. code-block:: bash

        salt '*' saltutil.sync_serializers
        salt '*' saltutil.sync_serializers saltenv=dev
        salt '*' saltutil.sync_serializers saltenv=base,dev
    """
    ret = _sync("serializers", saltenv, extmod_whitelist, extmod_blacklist)
    if refresh:
        refresh_modules()
    return ret


def list_extmods():
    """
    .. versionadded:: 2017.7.0

    List Salt modules which have been synced externally

    CLI Examples:

    .. code-block:: bash

        salt '*' saltutil.list_extmods
    """
    ret = {}
    ext_dir = os.path.join(__opts__["cachedir"], "extmods")
    mod_types = os.listdir(ext_dir)
    for mod_type in mod_types:
        ret[mod_type] = set()
        for _, _, files in salt.utils.path.os_walk(os.path.join(ext_dir, mod_type)):
            for fh_ in files:
                ret[mod_type].add(fh_.split(".")[0])
        ret[mod_type] = list(ret[mod_type])
    return ret


def sync_log_handlers(
    saltenv=None, refresh=True, extmod_whitelist=None, extmod_blacklist=None
):
    """
    .. versionadded:: 2015.8.0

    Sync log handlers from ``salt://_log_handlers`` to the minion

    saltenv
        The fileserver environment from which to sync. To sync from more than
        one environment, pass a comma-separated list.

        If not passed, then all environments configured in the :ref:`top files
        <states-top>` will be checked for log handlers to sync. If no top files
        are found, then the ``base`` environment will be synced.

    refresh : True
        If ``True``, refresh the available execution modules on the minion.
        This refresh will be performed even if no new log handlers are synced.
        Set to ``False`` to prevent this refresh.

    extmod_whitelist : None
        comma-separated list of modules to sync

    extmod_blacklist : None
        comma-separated list of modules to blacklist based on type

    CLI Examples:

    .. code-block:: bash

        salt '*' saltutil.sync_log_handlers
        salt '*' saltutil.sync_log_handlers saltenv=dev
        salt '*' saltutil.sync_log_handlers saltenv=base,dev
    """
    ret = _sync("log_handlers", saltenv, extmod_whitelist, extmod_blacklist)
    if refresh:
        refresh_modules()
    return ret


def sync_pillar(
    saltenv=None,
    refresh=True,
    extmod_whitelist=None,
    extmod_blacklist=None,
    clean_pillar_cache=False,
):
    """
    .. versionadded:: 2015.8.11,2016.3.2

    Sync pillar modules from the ``salt://_pillar`` directory on the Salt
    fileserver. This function is environment-aware, pass the desired
    environment to grab the contents of the ``_pillar`` directory from that
    environment. The default environment, if none is specified,  is ``base``.

    refresh : True
        Also refresh the execution modules available to the minion, and refresh
        pillar data.

    extmod_whitelist : None
        comma-separated list of modules to sync

    extmod_blacklist : None
        comma-separated list of modules to blacklist based on type

    clean_pillar_cache : False
        Set to ``True`` to refresh pillar cache.

    .. note::
        This function will raise an error if executed on a traditional (i.e.
        not masterless) minion

    CLI Examples:

    .. code-block:: bash

        salt '*' saltutil.sync_pillar
        salt '*' saltutil.sync_pillar saltenv=dev
    """
    if __opts__["file_client"] != "local":
        raise CommandExecutionError(
            "Pillar modules can only be synced to masterless minions"
        )
    ret = _sync("pillar", saltenv, extmod_whitelist, extmod_blacklist)
    if refresh:
        # we don't need to call refresh_modules here because it's done by refresh_pillar
        refresh_pillar(clean_cache=clean_pillar_cache)
    return ret


def sync_tops(
    saltenv=None,
    refresh=True,
    extmod_whitelist=None,
    extmod_blacklist=None,
):
    """
    .. versionadded:: 3007.0

    Sync master tops from ``salt://_tops`` to the minion.

    saltenv
        The fileserver environment from which to sync. To sync from more than
        one environment, pass a comma-separated list.

        If not passed, then all environments configured in the :ref:`top files
        <states-top>` will be checked for master tops to sync. If no top files
        are found, then the ``base`` environment will be synced.

    refresh : True
        Refresh this module's cache containing the environments from which
        extension modules are synced when ``saltenv`` is not specified.
        This refresh will be performed even if no new master tops are synced.
        Set to ``False`` to prevent this refresh.

    extmod_whitelist : None
        comma-separated list of modules to sync

    extmod_blacklist : None
        comma-separated list of modules to blacklist based on type

    .. note::
        This function will raise an error if executed on a traditional (i.e.
        not masterless) minion

    CLI Examples:

    .. code-block:: bash

        salt '*' saltutil.sync_tops
        salt '*' saltutil.sync_tops saltenv=dev
    """
    if __opts__["file_client"] != "local":
        raise CommandExecutionError(
            "Master top modules can only be synced to masterless minions"
        )
    if refresh:
        __context__.pop(TOP_ENVS_CKEY, None)
    return _sync("tops", saltenv, extmod_whitelist, extmod_blacklist)


def sync_executors(
    saltenv=None, refresh=True, extmod_whitelist=None, extmod_blacklist=None
):
    """
    .. versionadded:: 3000

    Sync executors from ``salt://_executors`` to the minion

    saltenv
        The fileserver environment from which to sync. To sync from more than
        one environment, pass a comma-separated list.

        If not passed, then all environments configured in the :ref:`top files
        <states-top>` will be checked for log handlers to sync. If no top files
        are found, then the ``base`` environment will be synced.

    refresh : True
        If ``True``, refresh the available execution modules on the minion.
        This refresh will be performed even if no new log handlers are synced.
        Set to ``False`` to prevent this refresh.

    extmod_whitelist : None
        comma-seperated list of modules to sync

    extmod_blacklist : None
        comma-seperated list of modules to blacklist based on type

    CLI Examples:

    .. code-block:: bash

        salt '*' saltutil.sync_executors
        salt '*' saltutil.sync_executors saltenv=dev
        salt '*' saltutil.sync_executors saltenv=base,dev
    """
    ret = _sync("executors", saltenv, extmod_whitelist, extmod_blacklist)
    if refresh:
        refresh_modules()
    return ret


def sync_wrapper(
    saltenv=None, refresh=True, extmod_whitelist=None, extmod_blacklist=None
):
    """
    .. versionadded:: 3007.0

    Sync salt-ssh wrapper modules from ``salt://_wrapper`` to the minion.

    saltenv
        The fileserver environment from which to sync. To sync from more than
        one environment, pass a comma-separated list.

        If not passed, then all environments configured in the :ref:`top files
        <states-top>` will be checked for wrappers to sync. If no top files
        are found, then the ``base`` environment will be synced.

    refresh : True
        If ``True``, refresh the available wrapper modules on the minion.
        This refresh will be performed even if no wrappers are synced.
        Set to ``False`` to prevent this refresh.

    extmod_whitelist : None
        comma-seperated list of modules to sync

    extmod_blacklist : None
        comma-seperated list of modules to blacklist based on type

    .. note::
        This function will raise an error if executed on a traditional (i.e.
        not masterless) minion.

    CLI Examples:

    .. code-block:: bash

        salt '*' saltutil.sync_wrapper
        salt '*' saltutil.sync_wrapper saltenv=dev
        salt '*' saltutil.sync_wrapper saltenv=base,dev
    """
    if __opts__["file_client"] != "local":
        raise CommandExecutionError(
            "Wrapper modules can only be synced to masterless minions"
        )
    ret = _sync("wrapper", saltenv, extmod_whitelist, extmod_blacklist)
    if refresh:
        refresh_modules()
    return ret


def sync_all(
    saltenv=None,
    refresh=True,
    extmod_whitelist=None,
    extmod_blacklist=None,
    clean_pillar_cache=False,
):
    """
    .. versionchanged:: 3007.0

        On masterless minions, master top modules are now synced as well.
        When ``refresh`` is set to ``True``, this module's cache containing
        the environments from which extension modules are synced when
        ``saltenv`` is not specified will be refreshed.

    .. versionchanged:: 2015.8.11,2016.3.2
        On masterless minions, pillar modules are now synced, and refreshed
        when ``refresh`` is set to ``True``.

    Sync down all of the dynamic modules from the file server for a specific
    environment. This function synchronizes custom modules, states, beacons,
    grains, returners, output modules, renderers, and utils.

    refresh : True
        Also refresh the execution modules and recompile pillar data available
        to the minion. If this is a masterless minion, also refresh the environments
        from which extension modules are synced after syncing master tops.
        This refresh will be performed even if no new dynamic
        modules are synced. Set to ``False`` to prevent this refresh.

    .. important::

        If this function is executed using a :py:func:`module.run
        <salt.states.module.run>` state, the SLS file will not have access to
        newly synced execution modules unless a ``refresh`` argument is
        added to the state, like so:

        .. code-block:: yaml

            load_my_custom_module:
              module.run:
                - name: saltutil.sync_all
                - refresh: True

        See :ref:`here <reloading-modules>` for a more detailed explanation of
        why this is necessary.

    extmod_whitelist : None
        dictionary of modules to sync based on type

    extmod_blacklist : None
        dictionary of modules to blacklist based on type

    clean_pillar_cache : False
        Set to ``True`` to refresh pillar cache.

    CLI Examples:

    .. code-block:: bash

        salt '*' saltutil.sync_all
        salt '*' saltutil.sync_all saltenv=dev
        salt '*' saltutil.sync_all saltenv=base,dev
        salt '*' saltutil.sync_all extmod_whitelist={'modules': ['custom_module']}
    """
    log.debug("Syncing all")
    ret = {}
    if __opts__["file_client"] == "local":
        # Sync tops first since this might influence the other syncs
        ret["tops"] = sync_tops(saltenv, refresh, extmod_whitelist, extmod_blacklist)
    ret["clouds"] = sync_clouds(saltenv, False, extmod_whitelist, extmod_blacklist)
    ret["beacons"] = sync_beacons(saltenv, False, extmod_whitelist, extmod_blacklist)
    ret["modules"] = sync_modules(saltenv, False, extmod_whitelist, extmod_blacklist)
    ret["states"] = sync_states(saltenv, False, extmod_whitelist, extmod_blacklist)
    ret["sdb"] = sync_sdb(saltenv, extmod_whitelist, extmod_blacklist)
    ret["grains"] = sync_grains(saltenv, False, extmod_whitelist, extmod_blacklist)
    ret["renderers"] = sync_renderers(
        saltenv, False, extmod_whitelist, extmod_blacklist
    )
    ret["returners"] = sync_returners(
        saltenv, False, extmod_whitelist, extmod_blacklist
    )
    ret["output"] = sync_output(saltenv, False, extmod_whitelist, extmod_blacklist)
    ret["utils"] = sync_utils(saltenv, False, extmod_whitelist, extmod_blacklist)
    ret["log_handlers"] = sync_log_handlers(
        saltenv, False, extmod_whitelist, extmod_blacklist
    )
    ret["executors"] = sync_executors(
        saltenv, False, extmod_whitelist, extmod_blacklist
    )
    ret["proxymodules"] = sync_proxymodules(
        saltenv, False, extmod_whitelist, extmod_blacklist
    )
    ret["engines"] = sync_engines(saltenv, False, extmod_whitelist, extmod_blacklist)
    ret["thorium"] = sync_thorium(saltenv, False, extmod_whitelist, extmod_blacklist)
    ret["serializers"] = sync_serializers(
        saltenv, False, extmod_whitelist, extmod_blacklist
    )
    ret["matchers"] = sync_matchers(saltenv, False, extmod_whitelist, extmod_blacklist)
    if __opts__["file_client"] == "local":
        ret["pillar"] = sync_pillar(saltenv, False, extmod_whitelist, extmod_blacklist)
        ret["wrapper"] = sync_wrapper(
            saltenv, False, extmod_whitelist, extmod_blacklist
        )
    if refresh:
        # we don't need to call refresh_modules here because it's done by refresh_pillar
        refresh_pillar(clean_cache=clean_pillar_cache)
    return ret


def refresh_beacons():
    """
    Signal the minion to refresh the beacons.

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.refresh_beacons
    """
    try:
        ret = __salt__["event.fire"]({}, "beacons_refresh")
    except KeyError:
        log.error("Event module not available. Module refresh failed.")
        ret = False  # Effectively a no-op, since we can't really return without an event system
    return ret


def refresh_matchers():
    """
    Signal the minion to refresh its matchers.

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.refresh_matchers
    """
    try:
        ret = __salt__["event.fire"]({}, "matchers_refresh")
    except KeyError:
        log.error("Event module not available. Matcher refresh failed.")
        ret = False  # Effectively a no-op, since we can't really return without an event system
    return ret


def refresh_pillar(wait=False, timeout=30, clean_cache=True):
    """
    Signal the minion to refresh the in-memory pillar data. See :ref:`pillar-in-memory`.

    :param wait:            Wait for pillar refresh to complete, defaults to False.
    :type wait:             bool, optional
    :param timeout:         How long to wait in seconds, only used when wait is True, defaults to 30.
    :type timeout:          int, optional
    :param clean_cache:     Clean the pillar cache, only used when `pillar_cache` is True. Defaults to True
    :type clean_cache:      bool, optional
        .. versionadded:: 3005
    :return:                Boolean status, True when the pillar_refresh event was fired successfully.

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.refresh_pillar
        salt '*' saltutil.refresh_pillar wait=True timeout=60
    """
    data = {"clean_cache": clean_cache}
    try:
        if wait:
            #  If we're going to block, first setup a listener
            with salt.utils.event.get_event(
                "minion", opts=__opts__, listen=True
            ) as eventer:
                ret = __salt__["event.fire"](data, "pillar_refresh")
                # Wait for the finish event to fire
                log.trace("refresh_pillar waiting for pillar refresh to complete")
                # Blocks until we hear this event or until the timeout expires
                event_ret = eventer.get_event(
                    tag=salt.defaults.events.MINION_PILLAR_REFRESH_COMPLETE,
                    wait=timeout,
                )
                if not event_ret or event_ret["complete"] is False:
                    log.warning(
                        "Pillar refresh did not complete within timeout %s", timeout
                    )
        else:
            ret = __salt__["event.fire"](data, "pillar_refresh")
    except KeyError:
        log.error("Event module not available. Pillar refresh failed.")
        ret = False  # Effectively a no-op, since we can't really return without an event system
    return ret


pillar_refresh = salt.utils.functools.alias_function(refresh_pillar, "pillar_refresh")


def refresh_modules(**kwargs):
    """
    Signal the minion to refresh the module and grain data

    The default is to refresh module asynchronously. To block
    until the module refresh is complete, set the 'async' flag
    to False.

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.refresh_modules
    """
    asynchronous = bool(kwargs.get("async", True))
    try:
        if asynchronous:
            ret = __salt__["event.fire"]({}, "module_refresh")
        else:
            #  If we're going to block, first setup a listener
            with salt.utils.event.get_event(
                "minion", opts=__opts__, listen=True
            ) as eventer:
                ret = __salt__["event.fire"]({"notify": True}, "module_refresh")
                # Wait for the finish event to fire
                log.trace("refresh_modules waiting for module refresh to complete")
                # Blocks until we hear this event or until the timeout expires
                eventer.get_event(
                    tag=salt.defaults.events.MINION_MOD_REFRESH_COMPLETE, wait=30
                )
    except KeyError:
        log.error("Event module not available. Module refresh failed.")
        ret = False  # Effectively a no-op, since we can't really return without an event system
    return ret


def is_running(fun):
    """
    If the named function is running return the data associated with it/them.
    The argument can be a glob

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.is_running state.highstate
    """
    run = running()
    ret = []
    for data in run:
        if fnmatch.fnmatch(data.get("fun", ""), fun):
            ret.append(data)
    return ret


def running():
    """
    Return the data on all running salt processes on the minion

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.running
    """
    return salt.utils.minion.running(__opts__)


def clear_cache():
    """
    Forcibly removes all caches on a minion.

    .. versionadded:: 2014.7.0

    WARNING: The safest way to clear a minion cache is by first stopping
    the minion and then deleting the cache files before restarting it.

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.clear_cache
    """
    for root, dirs, files in salt.utils.files.safe_walk(
        __opts__["cachedir"], followlinks=False
    ):
        for name in files:
            try:
                os.remove(os.path.join(root, name))
            except OSError as exc:
                log.error(
                    "Attempt to clear cache with saltutil.clear_cache FAILED with: %s",
                    exc,
                )
                return False
    return True


def clear_job_cache(hours=24):
    """
    Forcibly removes job cache folders and files on a minion.

    .. versionadded:: 2018.3.0

    WARNING: The safest way to clear a minion cache is by first stopping
    the minion and then deleting the cache files before restarting it.

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.clear_job_cache hours=12
    """
    threshold = time.time() - hours * 3600
    for root, dirs, files in salt.utils.files.safe_walk(
        os.path.join(__opts__["cachedir"], "minion_jobs"), followlinks=False
    ):
        for name in dirs:
            try:
                directory = os.path.join(root, name)
                mtime = os.path.getmtime(directory)
                if mtime < threshold:
                    shutil.rmtree(directory)
            except OSError as exc:
                log.error(
                    "Attempt to clear cache with saltutil.clear_job_cache FAILED"
                    " with: %s",
                    exc,
                )
                return False
    return True


def find_job(jid):
    """
    Return the data for a specific job id that is currently running.

    jid
        The job id to search for and return data.

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.find_job <job id>

    Note that the find_job function only returns job information when the job is still running. If
    the job is currently running, the output looks something like this:

    .. code-block:: bash

        # salt my-minion saltutil.find_job 20160503150049487736
        my-minion:
            ----------
            arg:
                - 30
            fun:
                test.sleep
            jid:
                20160503150049487736
            pid:
                9601
            ret:
            tgt:
                my-minion
            tgt_type:
                glob
            user:
                root

    If the job has already completed, the job cannot be found and therefore the function returns
    an empty dictionary, which looks like this on the CLI:

    .. code-block:: bash

        # salt my-minion saltutil.find_job 20160503150049487736
        my-minion:
            ----------
    """
    for data in running():
        if data["jid"] == jid:
            return data
    return {}


def find_cached_job(jid):
    """
    Return the data for a specific cached job id. Note this only works if
    cache_jobs has previously been set to True on the minion.

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.find_cached_job <job id>
    """
    proc_dir = os.path.join(__opts__["cachedir"], "minion_jobs")
    job_dir = os.path.join(proc_dir, str(jid))
    if not os.path.isdir(job_dir):
        if not __opts__.get("cache_jobs"):
            return (
                "Local jobs cache directory not found; you may need to"
                " enable cache_jobs on this minion"
            )
        else:
            return f"Local jobs cache directory {job_dir} not found"
    path = os.path.join(job_dir, "return.p")
    with salt.utils.files.fopen(path, "rb") as fp_:
        buf = fp_.read()
    if buf:
        try:
            data = salt.payload.loads(buf)
        except NameError:
            # msgpack error in salt-ssh
            pass
        else:
            if isinstance(data, dict):
                # if not a dict, this was an invalid serialized object
                return data
    return None


def signal_job(jid, sig):
    """
    Sends a signal to the named salt job's process

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.signal_job <job id> 15
    """
    if HAS_PSUTIL is False:
        log.warning(
            "saltutil.signal job called, but psutil is not installed. "
            "Install psutil to ensure more reliable and accurate PID "
            "management."
        )
    for data in running():
        if data["jid"] == jid:
            try:
                if HAS_PSUTIL:
                    for proc in psutil.Process(pid=data["pid"]).children(
                        recursive=True
                    ):
                        proc.send_signal(sig)
                os.kill(int(data["pid"]), sig)
                if HAS_PSUTIL is False and "child_pids" in data:
                    for pid in data["child_pids"]:
                        os.kill(int(pid), sig)
                return "Signal {} sent to job {} at pid {}".format(
                    int(sig), jid, data["pid"]
                )
            except OSError:
                path = os.path.join(__opts__["cachedir"], "proc", str(jid))
                if os.path.isfile(path):
                    os.remove(path)
                return "Job {} was not running and job data has been cleaned up".format(
                    jid
                )
    return ""


def term_job(jid):
    """
    Sends a termination signal (SIGTERM 15) to the named salt job's process

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.term_job <job id>
    """
    return signal_job(jid, signal.SIGTERM)


def term_all_jobs():
    """
    Sends a termination signal (SIGTERM 15) to all currently running jobs

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.term_all_jobs
    """
    ret = []
    for data in running():
        ret.append(signal_job(data["jid"], signal.SIGTERM))
    return ret


def kill_job(jid):
    """
    Sends a kill signal (SIGKILL 9) to the named salt job's process

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.kill_job <job id>
    """
    # Some OS's (Win32) don't have SIGKILL, so use salt_SIGKILL which is set to
    # an appropriate value for the operating system this is running on.
    return signal_job(jid, salt_SIGKILL)


def kill_all_jobs():
    """
    Sends a kill signal (SIGKILL 9) to all currently running jobs

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.kill_all_jobs
    """
    # Some OS's (Win32) don't have SIGKILL, so use salt_SIGKILL which is set to
    # an appropriate value for the operating system this is running on.
    ret = []
    for data in running():
        ret.append(signal_job(data["jid"], salt_SIGKILL))
    return ret


def regen_keys():
    """
    Used to regenerate the minion keys.

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.regen_keys
    """
    for fn_ in os.listdir(__opts__["pki_dir"]):
        path = os.path.join(__opts__["pki_dir"], fn_)
        try:
            os.remove(path)
        except OSError:
            pass
    # TODO: move this into a channel function? Or auth?
    # create a channel again, this will force the key regen
    with salt.channel.client.ReqChannel.factory(__opts__) as channel:
        log.debug("Recreating channel to force key regen")


def revoke_auth(preserve_minion_cache=False):
    """
    The minion sends a request to the master to revoke its own key.
    Note that the minion session will be revoked and the minion may
    not be able to return the result of this command back to the master.

    If the 'preserve_minion_cache' flag is set to True, the master
    cache for this minion will not be removed.

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.revoke_auth
    """
    masters = list()
    ret = True
    if "master_uri_list" in __opts__:
        for master_uri in __opts__["master_uri_list"]:
            masters.append(master_uri)
    else:
        masters.append(__opts__["master_uri"])

    for master in masters:
        with salt.channel.client.ReqChannel.factory(
            __opts__, master_uri=master
        ) as channel:
            tok = channel.auth.gen_token(b"salt")
            load = {
                "cmd": "revoke_auth",
                "id": __opts__["id"],
                "tok": tok,
                "preserve_minion_cache": preserve_minion_cache,
            }
            try:
                channel.send(load)
            except SaltReqTimeoutError:
                ret = False
    return ret


def _get_ssh_or_api_client(cfgfile, ssh=False):
    if ssh:
        client = salt.client.ssh.client.SSHClient(cfgfile)
    else:
        client = salt.client.get_local_client(cfgfile)
    return client


def _exec(
    client,
    tgt,
    fun,
    arg,
    timeout,
    tgt_type,
    ret,
    kwarg,
    batch=False,
    subset=False,
    **kwargs,
):
    fcn_ret = {}
    seen = 0

    cmd_kwargs = {
        "tgt": tgt,
        "fun": fun,
        "arg": arg,
        "timeout": timeout,
        "tgt_type": tgt_type,
        "ret": ret,
        "kwarg": kwarg,
    }

    if batch:
        _cmd = client.cmd_batch
        cmd_kwargs.update({"batch": batch})
    elif subset:
        _cmd = client.cmd_subset
        cmd_kwargs.update({"subset": subset, "cli": True})
    else:
        _cmd = client.cmd_iter

    cmd_kwargs.update(kwargs)
    for ret_comp in _cmd(**cmd_kwargs):
        fcn_ret.update(ret_comp)
        seen += 1
        # fcn_ret can be empty, so we cannot len the whole return dict
        if tgt_type == "list" and len(tgt) == seen:
            # do not wait for timeout when explicit list matching
            # and all results are there
            break

    if batch:
        old_ret, fcn_ret = fcn_ret, {}
        for key, value in old_ret.items():
            fcn_ret[key] = {
                "out": (
                    value.get("out", "highstate")
                    if isinstance(value, dict)
                    else "highstate"
                ),
                "ret": value,
            }

    return fcn_ret


def cmd(
    tgt,
    fun,
    arg=(),
    timeout=None,
    tgt_type="glob",
    ret="",
    kwarg=None,
    ssh=False,
    **kwargs,
):
    """
    .. versionchanged:: 2017.7.0
        The ``expr_form`` argument has been renamed to ``tgt_type``, earlier
        releases must use ``expr_form``.

    Assuming this minion is a master, execute a salt command

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.cmd
    """
    cfgfile = __opts__["conf_file"]
    with _get_ssh_or_api_client(cfgfile, ssh) as client:
        fcn_ret = _exec(client, tgt, fun, arg, timeout, tgt_type, ret, kwarg, **kwargs)
    # if return is empty, we may have not used the right conf,
    # try with the 'minion relative master configuration counter part
    # if available
    master_cfgfile = f"{cfgfile[:-6]}master"  # remove 'minion'
    if (
        not fcn_ret
        and cfgfile.endswith("{}{}".format(os.path.sep, "minion"))
        and os.path.exists(master_cfgfile)
    ):
        with _get_ssh_or_api_client(master_cfgfile, ssh) as client:
            fcn_ret = _exec(
                client, tgt, fun, arg, timeout, tgt_type, ret, kwarg, **kwargs
            )

    return fcn_ret


def cmd_iter(
    tgt,
    fun,
    arg=(),
    timeout=None,
    tgt_type="glob",
    ret="",
    kwarg=None,
    ssh=False,
    **kwargs,
):
    """
    .. versionchanged:: 2017.7.0
        The ``expr_form`` argument has been renamed to ``tgt_type``, earlier
        releases must use ``expr_form``.

    Assuming this minion is a master, execute a salt command

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.cmd_iter
    """
    if ssh:
        client = salt.client.ssh.client.SSHClient(__opts__["conf_file"])
    else:
        client = salt.client.get_local_client(__opts__["conf_file"])
    for ret in client.cmd_iter(tgt, fun, arg, timeout, tgt_type, ret, kwarg, **kwargs):
        yield ret


def runner(
    name, arg=None, kwarg=None, full_return=False, saltenv="base", jid=None, **kwargs
):
    """
    Execute a runner function. This function must be run on the master,
    either by targeting a minion running on a master or by using
    salt-call on a master.

    .. versionadded:: 2014.7.0

    name
        The name of the function to run

    kwargs
        Any keyword arguments to pass to the runner function

    CLI Example:

    In this example, assume that `master_minion` is a minion running
    on a master.

    .. code-block:: bash

        salt master_minion saltutil.runner jobs.list_jobs
        salt master_minion saltutil.runner test.arg arg="['baz']" kwarg="{'foo': 'bar'}"
    """
    if arg is None:
        arg = []
    if kwarg is None:
        kwarg = {}
    pub_data = {}
    jid = kwargs.pop("__orchestration_jid__", jid)
    saltenv = kwargs.pop("__env__", saltenv)
    pub_data["user"] = kwargs.pop("__pub_user", "UNKNOWN")
    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    if kwargs:
        kwarg.update(kwargs)

    if "master_job_cache" not in __opts__:
        master_config = os.path.join(os.path.dirname(__opts__["conf_file"]), "master")
        master_opts = salt.config.master_config(master_config)
        rclient = salt.runner.RunnerClient(master_opts)
    else:
        rclient = salt.runner.RunnerClient(__opts__)

    if name in rclient.functions:
        aspec = salt.utils.args.get_function_argspec(rclient.functions[name])
        if "saltenv" in aspec.args:
            kwarg["saltenv"] = saltenv

    if name in ["state.orchestrate", "state.orch", "state.sls"]:
        kwarg["orchestration_jid"] = jid

    if jid:
        salt.utils.event.fire_args(
            __opts__,
            jid,
            {"type": "runner", "name": name, "args": arg, "kwargs": kwarg},
            prefix="run",
        )

    return rclient.cmd(
        name,
        arg=arg,
        pub_data=pub_data,
        kwarg=kwarg,
        print_event=False,
        full_return=full_return,
    )


def wheel(name, *args, **kwargs):
    """
    Execute a wheel module and function. This function must be run against a
    minion that is local to the master.

    .. versionadded:: 2014.7.0

    name
        The name of the function to run

    args
        Any positional arguments to pass to the wheel function. A common example
        of this would be the ``match`` arg needed for key functions.

        .. versionadded:: 2015.8.11

    kwargs
        Any keyword arguments to pass to the wheel function

    CLI Example:

    .. code-block:: bash

        salt my-local-minion saltutil.wheel key.accept jerry
        salt my-local-minion saltutil.wheel minions.connected

    .. note::

        Since this function must be run against a minion that is running locally
        on the master in order to get accurate returns, if this function is run
        against minions that are not local to the master, "empty" returns are
        expected. The remote minion does not have access to wheel functions and
        their return data.

    """
    jid = kwargs.pop("__orchestration_jid__", None)
    saltenv = kwargs.pop("__env__", "base")

    if __opts__["__role"] == "minion":
        master_config = os.path.join(os.path.dirname(__opts__["conf_file"]), "master")
        master_opts = salt.config.client_config(master_config)
        wheel_client = salt.wheel.WheelClient(master_opts)
    else:
        wheel_client = salt.wheel.WheelClient(__opts__)

    # The WheelClient cmd needs args, kwargs, and pub_data separated out from
    # the "normal" kwargs structure, which at this point contains __pub_x keys.
    pub_data = {}
    valid_kwargs = {}
    for key, val in kwargs.items():
        if key.startswith("__"):
            pub_data[key] = val
        else:
            valid_kwargs[key] = val

    try:
        if name in wheel_client.functions:
            aspec = salt.utils.args.get_function_argspec(wheel_client.functions[name])
            if "saltenv" in aspec.args:
                valid_kwargs["saltenv"] = saltenv

        if jid:
            salt.utils.event.fire_args(
                __opts__,
                jid,
                {"type": "wheel", "name": name, "args": valid_kwargs},
                prefix="run",
            )

        ret = wheel_client.cmd(
            name,
            arg=args,
            pub_data=pub_data,
            kwarg=valid_kwargs,
            print_event=False,
            full_return=True,
        )
    except SaltInvocationError:
        raise CommandExecutionError(
            "This command can only be executed on a minion that is located on "
            "the master."
        )

    return ret


# this is the only way I could figure out how to get the REAL file_roots
# __opt__['file_roots'] is set to  __opt__['pillar_root']
class _MMinion:
    def __new__(cls, saltenv, reload_env=False):
        # this is to break out of salt.loaded.int and make this a true singleton
        # hack until https://github.com/saltstack/salt/pull/10273 is resolved
        # this is starting to look like PHP
        global _mminions  # pylint: disable=W0601
        if "_mminions" not in globals():
            _mminions = {}
        if saltenv not in _mminions or reload_env:
            opts = copy.deepcopy(__opts__)
            del opts["file_roots"]
            # grains at this point are in the context of the minion
            global __grains__  # pylint: disable=W0601
            grains = copy.deepcopy(__grains__)
            m = salt.minion.MasterMinion(opts)

            # this assignment is so that the rest of fxns called by salt still
            # have minion context
            __grains__ = grains

            # this assignment is so that fxns called by mminion have minion
            # context
            m.opts["grains"] = grains

            env_roots = m.opts["file_roots"][saltenv]
            m.opts["module_dirs"] = [fp + "/_modules" for fp in env_roots]
            m.gen_modules()
            _mminions[saltenv] = m
        return _mminions[saltenv]


def mmodule(saltenv, fun, *args, **kwargs):
    """
    Loads minion modules from an environment so that they can be used in pillars
    for that environment

    CLI Example:

    .. code-block:: bash

        salt '*' saltutil.mmodule base test.ping
    """
    mminion = _MMinion(saltenv)
    return mminion.functions[fun](*args, **kwargs)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.drbd
"""

import pytest

import salt.modules.drbd as drbd
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {drbd: {}}


# 'overview' function tests: 1


def test_overview():
    """
    Test if it shows status of the DRBD devices
    """
    ret = {
        "connection state": "True",
        "device": "Stack",
        "fs": "None",
        "local disk state": "UpToDate",
        "local role": "master",
        "minor number": "Salt",
        "mountpoint": "True",
        "partner disk state": "UpToDate",
        "partner role": "minion",
        "percent": "888",
        "remains": "666",
        "total size": "50",
        "used": "50",
    }
    mock = MagicMock(
        return_value=(
            "Salt:Stack True master/minion UpToDate/UpToDate True None 50 50 666 888"
        )
    )
    with patch.dict(drbd.__salt__, {"cmd.run": mock}):
        assert drbd.overview() == ret

    ret = {
        "connection state": "True",
        "device": "Stack",
        "local disk state": "UpToDate",
        "local role": "master",
        "minor number": "Salt",
        "partner disk state": "partner",
        "partner role": "minion",
        "synched": "5050",
        "synchronisation: ": "syncbar",
    }
    mock = MagicMock(
        return_value=(
            "Salt:Stack True master/minion UpToDate/partner syncbar None 50 50"
        )
    )
    with patch.dict(drbd.__salt__, {"cmd.run": mock}):
        assert drbd.overview() == ret


def test_status():
    """
    Test if it shows status of the DRBD resources via drbdadm
    """
    ret = [
        {
            "local role": "Primary",
            "local volumes": [{"disk": "UpToDate"}],
            "peer nodes": [
                {
                    "peer volumes": [
                        {
                            "done": "96.47",
                            "peer-disk": "Inconsistent",
                            "replication": "SyncSource",
                        }
                    ],
                    "peernode name": "opensuse-node2",
                    "role": "Secondary",
                }
            ],
            "resource name": "single",
        }
    ]

    mock = MagicMock(
        return_value="""
single role:Primary
  disk:UpToDate
  opensuse-node2 role:Secondary
    replication:SyncSource peer-disk:Inconsistent done:96.47
"""
    )

    with patch.dict(drbd.__salt__, {"cmd.run": mock}):
        assert drbd.status() == ret

    ret = [
        {
            "local role": "Primary",
            "local volumes": [
                {"disk": "UpToDate", "volume": "0"},
                {"disk": "UpToDate", "volume": "1"},
            ],
            "peer nodes": [
                {
                    "peer volumes": [
                        {"peer-disk": "UpToDate", "volume": "0"},
                        {"peer-disk": "UpToDate", "volume": "1"},
                    ],
                    "peernode name": "node2",
                    "role": "Secondary",
                },
                {
                    "peer volumes": [
                        {"peer-disk": "UpToDate", "volume": "0"},
                        {"peer-disk": "UpToDate", "volume": "1"},
                    ],
                    "peernode name": "node3",
                    "role": "Secondary",
                },
            ],
            "resource name": "res",
        },
        {
            "local role": "Primary",
            "local volumes": [
                {"disk": "UpToDate", "volume": "0"},
                {"disk": "UpToDate", "volume": "1"},
            ],
            "peer nodes": [
                {
                    "peer volumes": [
                        {"peer-disk": "UpToDate", "volume": "0"},
                        {"peer-disk": "UpToDate", "volume": "1"},
                    ],
                    "peernode name": "node2",
                    "role": "Secondary",
                },
                {
                    "peer volumes": [
                        {"peer-disk": "UpToDate", "volume": "0"},
                        {"peer-disk": "UpToDate", "volume": "1"},
                    ],
                    "peernode name": "node3",
                    "role": "Secondary",
                },
            ],
            "resource name": "test",
        },
    ]

    mock = MagicMock(
        return_value="""
res role:Primary
  volume:0 disk:UpToDate
  volume:1 disk:UpToDate
  node2 role:Secondary
    volume:0 peer-disk:UpToDate
    volume:1 peer-disk:UpToDate
  node3 role:Secondary
    volume:0 peer-disk:UpToDate
    volume:1 peer-disk:UpToDate

test role:Primary
  volume:0 disk:UpToDate
  volume:1 disk:UpToDate
  node2 role:Secondary
    volume:0 peer-disk:UpToDate
    volume:1 peer-disk:UpToDate
  node3 role:Secondary
    volume:0 peer-disk:UpToDate
    volume:1 peer-disk:UpToDate

"""
    )
    with patch.dict(drbd.__salt__, {"cmd.run": mock}):
        assert drbd.status() == ret



"""
Decorators for salt.utils.path
"""

import functools

import salt.utils.path
from salt.exceptions import CommandNotFoundError


def which(exe):
    """
    Decorator wrapper for salt.utils.path.which
    """

    def wrapper(function):
        @functools.wraps(function)
        def wrapped(*args, **kwargs):
            if salt.utils.path.which(exe) is None:
                raise CommandNotFoundError(
                    f"The '{exe}' binary was not found in $PATH."
                )
            return function(*args, **kwargs)

        return wrapped

    return wrapper


def which_bin(exes):
    """
    Decorator wrapper for salt.utils.path.which_bin
    """

    def wrapper(function):
        @functools.wraps(function)
        def wrapped(*args, **kwargs):
            if salt.utils.path.which_bin(exes) is None:
                raise CommandNotFoundError(
                    "None of provided binaries({}) were found in $PATH.".format(
                        [f"'{exe}'" for exe in exes]
                    )
                )
            return function(*args, **kwargs)

        return wrapped

    return wrapper



"""
An engine that sends events to the Logentries logging service.

:maintainer:  Jimmy Tang (jimmy_tang@rapid7.com)
:maturity:    New
:depends:     ssl, certifi
:platform:    all

.. versionadded:: 2016.3.0

To enable this engine the master and/or minion will need the following
python libraries

    ssl
    certifi

If you are running a new enough version of python then the ssl library
will be present already.

You will also need the following values configured in the minion or
master config.

:configuration:

    Example configuration

    .. code-block:: yaml

        engines:
          - logentries:
              endpoint: data.logentries.com
              port: 10000
              token: 057af3e2-1c05-47c5-882a-5cd644655dbf

The 'token' can be obtained from the Logentries service.

To test this engine

    .. code-block:: bash

         salt '*' test.ping cmd.run uptime

"""

import logging
import random
import socket
import time
import uuid

import salt.utils.event
import salt.utils.json

try:
    import certifi

    HAS_CERTIFI = True
except ImportError:
    HAS_CERTIFI = False

# This is here for older python installs, it is needed to setup an encrypted tcp connection
try:
    import ssl

    HAS_SSL = True
except ImportError:  # for systems without TLS support.
    HAS_SSL = False


log = logging.getLogger(__name__)


def __virtual__():
    return True if HAS_CERTIFI and HAS_SSL else False


class PlainTextSocketAppender:
    def __init__(
        self, verbose=True, LE_API="data.logentries.com", LE_PORT=80, LE_TLS_PORT=443
    ):

        self.LE_API = LE_API
        self.LE_PORT = LE_PORT
        self.LE_TLS_PORT = LE_TLS_PORT
        self.MIN_DELAY = 0.1
        self.MAX_DELAY = 10
        # Error message displayed when an incorrect Token has been detected
        self.INVALID_TOKEN = (
            "\n\nIt appears the LOGENTRIES_TOKEN "
            "parameter you entered is incorrect!\n\n"
        )
        # Encoded unicode line separator
        self.LINE_SEP = salt.utils.stringutils.to_str("\u2028")

        self.verbose = verbose
        self._conn = None

    def open_connection(self):
        self._conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._conn.connect((self.LE_API, self.LE_PORT))

    def reopen_connection(self):
        self.close_connection()

        root_delay = self.MIN_DELAY
        while True:
            try:
                self.open_connection()
                return
            except Exception:  # pylint: disable=broad-except
                if self.verbose:
                    log.warning("Unable to connect to Logentries")

            root_delay *= 2
            if root_delay > self.MAX_DELAY:
                root_delay = self.MAX_DELAY

            wait_for = root_delay + random.uniform(0, root_delay)

            try:
                time.sleep(wait_for)
            except KeyboardInterrupt:  # pylint: disable=try-except-raise
                raise

    def close_connection(self):
        if self._conn is not None:
            self._conn.close()

    def put(self, data):
        # Replace newlines with Unicode line separator for multi-line events
        multiline = data.replace("\n", self.LINE_SEP) + "\n"
        # Send data, reconnect if needed
        while True:
            try:
                self._conn.send(multiline)
            except OSError:
                self.reopen_connection()
                continue
            break

        self.close_connection()


try:
    import ssl

    HAS_SSL = True
except ImportError:  # for systems without TLS support.
    SocketAppender = PlainTextSocketAppender
    HAS_SSL = False
else:

    class TLSSocketAppender(PlainTextSocketAppender):
        def open_connection(self):
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock = ssl.wrap_socket(
                sock=sock,
                keyfile=None,
                certfile=None,
                server_side=False,
                cert_reqs=ssl.CERT_REQUIRED,
                ssl_version=getattr(ssl, "PROTOCOL_TLSv1_2", ssl.PROTOCOL_TLSv1),
                ca_certs=certifi.where(),
                do_handshake_on_connect=True,
                suppress_ragged_eofs=True,
            )
            sock.connect((self.LE_API, self.LE_TLS_PORT))
            self._conn = sock

    SocketAppender = TLSSocketAppender


def event_bus_context(opts):
    if opts.get("id").endswith("_master"):
        event_bus = salt.utils.event.get_master_event(
            opts, opts["sock_dir"], listen=True
        )
    else:
        event_bus = salt.utils.event.get_event(
            "minion",
            opts=opts,
            sock_dir=opts["sock_dir"],
            listen=True,
        )
    return event_bus


def start(
    endpoint="data.logentries.com",
    port=10000,
    token=None,
    tag="salt/engines/logentries",
):
    """
    Listen to salt events and forward them to Logentries
    """
    with event_bus_context(__opts__) as event_bus:
        log.debug("Logentries engine started")
        try:
            val = uuid.UUID(token)
        except ValueError:
            log.warning("Not a valid logentries token")

        appender = SocketAppender(verbose=False, LE_API=endpoint, LE_PORT=port)
        appender.reopen_connection()

        while True:
            event = event_bus.get_event()
            if event:
                msg = " ".join(
                    (
                        salt.utils.stringutils.to_str(token),
                        salt.utils.stringutils.to_str(tag),
                        salt.utils.json.dumps(event),
                    )
                )
                appender.put(msg)

        appender.close_connection()



"""
Manage groups on Windows

.. important::
    If you feel that Salt should be using this module to manage groups on a
    minion, and it is using a different module (or gives an error similar to
    *'group.info' is not available*), see :ref:`here
    <module-provider-override>`.
"""

import logging

import salt.utils.platform
import salt.utils.win_functions
import salt.utils.winapi

try:
    import pywintypes
    import win32api
    import win32com.client

    HAS_DEPENDENCIES = True
except ImportError:
    HAS_DEPENDENCIES = False

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "group"


def __virtual__():
    """
    Set the group module if the kernel is Windows
    """
    if not salt.utils.platform.is_windows():
        return False, "win_groupadd: only works on Windows systems"
    if not HAS_DEPENDENCIES:
        return False, "win_groupadd: missing dependencies"
    return __virtualname__


def _get_computer_object():
    """
    A helper function to get the object for the local machine

    Returns:
        object: Returns the computer object for the local machine
    """
    with salt.utils.winapi.Com():
        nt = win32com.client.Dispatch("AdsNameSpaces")
        return nt.GetObject("", "WinNT://.,computer")


def _get_group_object(name):
    """
    A helper function to get a specified group object

    Args:

        name (str): The name of the object

    Returns:
        object: The specified group object
    """
    with salt.utils.winapi.Com():
        nt = win32com.client.Dispatch("AdsNameSpaces")
        return nt.GetObject("", f"WinNT://./{name},group")


def _get_all_groups():
    """
    A helper function that gets a list of group objects for all groups on the
    machine

    Returns:
        iter: A list of objects for all groups on the machine
    """
    with salt.utils.winapi.Com():
        nt = win32com.client.Dispatch("AdsNameSpaces")
        results = nt.GetObject("", "WinNT://.")
        results.Filter = ["group"]
        return results


def _get_username(member):
    """
    Resolve the username from the member object returned from a group query

    Returns:
        str: The username converted to domain\\username format
    """
    return member.ADSPath.replace("WinNT://", "").replace("/", "\\")


def add(name, **kwargs):
    """
    Add the specified group

    Args:

        name (str):
            The name of the group to add

    Returns:
        bool: ``True`` if successful, otherwise ``False``

    CLI Example:

    .. code-block:: bash

        salt '*' group.add foo
    """
    if not info(name):
        comp_obj = _get_computer_object()
        try:
            new_group = comp_obj.Create("group", name)
            new_group.SetInfo()
            log.info("Successfully created group %s", name)
        except pywintypes.com_error as exc:
            log.error(
                "Failed to create group %s. %s",
                name,
                win32api.FormatMessage(exc.excepinfo[5]),
            )
            return False
    else:
        log.warning("The group %s already exists.", name)
        return False
    return True


def delete(name, **kwargs):
    """
    Remove the named group

    Args:

        name (str):
            The name of the group to remove

    Returns:
        bool: ``True`` if successful, otherwise ``False``

    CLI Example:

    .. code-block:: bash

        salt '*' group.delete foo
    """
    if info(name):
        comp_obj = _get_computer_object()
        try:
            comp_obj.Delete("group", name)
            log.info("Successfully removed group %s", name)
        except pywintypes.com_error as exc:
            log.error(
                "Failed to remove group %s. %s",
                name,
                win32api.FormatMessage(exc.excepinfo[5]),
            )
            return False
    else:
        log.warning("The group %s does not exist.", name)
        return False

    return True


def info(name):
    """
    Return information about a group

    Args:

        name (str):
            The name of the group for which to get information

    Returns:
        dict: A dictionary of information about the group

    CLI Example:

    .. code-block:: bash

        salt '*' group.info foo
    """
    try:
        groupObj = _get_group_object(name)
        gr_name = groupObj.Name
        gr_mem = [_get_username(x) for x in groupObj.members()]
    except pywintypes.com_error as exc:
        log.debug(
            "Failed to access group %s. %s",
            name,
            win32api.FormatMessage(exc.excepinfo[5]),
        )
        return False

    if not gr_name:
        return False

    return {"name": gr_name, "passwd": None, "gid": None, "members": gr_mem}


def getent(refresh=False):
    """
    Return info on all groups

    Args:

        refresh (bool):
            Refresh the info for all groups in ``__context__``. If False only
            the groups in ``__context__`` will be returned. If True the
            ``__context__`` will be refreshed with current data and returned.
            Default is False

    Returns:
        A list of groups and their information

    CLI Example:

    .. code-block:: bash

        salt '*' group.getent
    """
    if "group.getent" in __context__ and not refresh:
        return __context__["group.getent"]

    ret = []

    results = _get_all_groups()

    for result in results:
        group = {
            "gid": __salt__["file.group_to_gid"](result.Name),
            "members": [_get_username(x) for x in result.members()],
            "name": result.Name,
            "passwd": "x",
        }
        ret.append(group)
    __context__["group.getent"] = ret
    return ret


def adduser(name, username, **kwargs):
    """
    Add a user to a group

    Args:

        name (str):
            The name of the group to modify

        username (str):
            The name of the user to add to the group

    Returns:
        bool: ``True`` if successful, otherwise ``False``

    CLI Example:

    .. code-block:: bash

        salt '*' group.adduser foo username
    """
    try:
        group_obj = _get_group_object(name)
    except pywintypes.com_error as exc:
        log.error(
            "Failed to access group %s. %s",
            name,
            win32api.FormatMessage(exc.excepinfo[5]),
        )
        return False

    existing_members = [_get_username(x) for x in group_obj.members()]
    username = salt.utils.win_functions.get_sam_name(username)

    try:
        if username not in existing_members:
            group_obj.Add("WinNT://" + username.replace("\\", "/"))
            log.info("Added user %s", username)
        else:
            log.warning("User %s is already a member of %s", username, name)
            return False
    except pywintypes.com_error as exc:
        log.error("Failed to add %s to group %s. %s", username, name, exc.excepinfo[2])
        return False

    return True


def deluser(name, username, **kwargs):
    """
    Remove a user from a group

    Args:

        name (str):
            The name of the group to modify

        username (str):
            The name of the user to remove from the group

    Returns:
        bool: ``True`` if successful, otherwise ``False``

    CLI Example:

    .. code-block:: bash

        salt '*' group.deluser foo username
    """
    try:
        group_obj = _get_group_object(name)
    except pywintypes.com_error as exc:
        log.error(
            "Failed to access group %s. %s",
            name,
            win32api.FormatMessage(exc.excepinfo[5]),
        )
        return False

    existing_members = [_get_username(x) for x in group_obj.members()]

    try:
        if salt.utils.win_functions.get_sam_name(username) in existing_members:
            group_obj.Remove("WinNT://" + username.replace("\\", "/"))
            log.info("Removed user %s", username)
        else:
            log.warning("User %s is not a member of %s", username, name)
            return False
    except pywintypes.com_error as exc:
        log.error(
            "Failed to remove %s from group %s. %s",
            username,
            name,
            win32api.FormatMessage(exc.excepinfo[5]),
        )
        return False

    return True


def members(name, members_list, **kwargs):
    """
    Ensure a group contains only the members in the list

    Args:

        name (str):
            The name of the group to modify

        members_list (str):
            A single user or a comma separated list of users. The group will
            contain only the users specified in this list.

    Returns:
        bool: ``True`` if successful, otherwise ``False``

    CLI Example:

    .. code-block:: bash

        salt '*' group.members foo 'user1,user2,user3'
    """
    members_list = [
        salt.utils.win_functions.get_sam_name(m) for m in members_list.split(",")
    ]
    if not isinstance(members_list, list):
        log.debug("member_list is not a list")
        return False

    try:
        obj_group = _get_group_object(name)
    except pywintypes.com_error as exc:
        # Group probably doesn't exist, but we'll log the error
        log.error(
            "Failed to access group %s. %s",
            name,
            win32api.FormatMessage(exc.excepinfo[5]),
        )
        return False

    existing_members = [_get_username(x) for x in obj_group.members()]
    existing_members.sort()
    members_list.sort()

    if existing_members == members_list:
        log.info("%s membership is correct", name)
        return True

    # add users
    success = True
    for member in members_list:
        if member not in existing_members:
            try:
                obj_group.Add("WinNT://" + member.replace("\\", "/"))
                log.info("User added: %s", member)
            except pywintypes.com_error as exc:
                log.error(
                    "Failed to add %s to %s. %s",
                    member,
                    name,
                    win32api.FormatMessage(exc.excepinfo[5]),
                )
                success = False

    # remove users not in members_list
    for member in existing_members:
        if member not in members_list:
            try:
                obj_group.Remove("WinNT://" + member.replace("\\", "/"))
                log.info("User removed: %s", member)
            except pywintypes.com_error as exc:
                log.error(
                    "Failed to remove %s from %s. %s",
                    member,
                    name,
                    win32api.FormatMessage(exc.excepinfo[5]),
                )
                success = False

    return success


def list_groups(refresh=False):
    """
    Return a list of groups

    Args:

        refresh (bool):
            Refresh the info for all groups in ``__context__``. If False only
            the groups in ``__context__`` will be returned. If True, the
            ``__context__`` will be refreshed with current data and returned.
            Default is False

    Returns:
        list: A list of groups on the machine

    CLI Example:

    .. code-block:: bash

        salt '*' group.list_groups
    """
    if "group.list_groups" in __context__ and not refresh:
        return __context__["group.list_groups"]

    results = _get_all_groups()

    ret = []

    for result in results:
        ret.append(result.Name)

    __context__["group.list_groups"] = ret

    return ret



import pytest

import salt.states.grafana_datasource as grafana_datasource
from tests.support.mock import MagicMock, Mock, patch

profile = {
    "grafana_url": "http://grafana",
    "grafana_token": "token",
}


def mock_json_response(data):
    response = MagicMock()
    response.json = MagicMock(return_value=data)
    return Mock(return_value=response)


@pytest.fixture
def configure_loader_modules():
    return {grafana_datasource: {}}


def test_present():
    with patch("requests.get", mock_json_response([])):
        with patch("requests.post") as rpost:
            ret = grafana_datasource.present("test", "type", "url", profile=profile)
            rpost.assert_called_once_with(
                "http://grafana/api/datasources",
                grafana_datasource._get_json_data("test", "type", "url"),
                headers={
                    "Authorization": "Bearer token",
                    "Accept": "application/json",
                },
                timeout=3,
            )
            assert ret["result"]
            assert ret["comment"] == "New data source test added"

    data = grafana_datasource._get_json_data("test", "type", "url")
    data.update({"id": 1, "orgId": 1})
    with patch("requests.get", mock_json_response([data])):
        with patch("requests.put") as rput:
            ret = grafana_datasource.present("test", "type", "url", profile=profile)
            rput.assert_called_once_with(
                "http://grafana/api/datasources/1",
                grafana_datasource._get_json_data("test", "type", "url"),
                headers={
                    "Authorization": "Bearer token",
                    "Accept": "application/json",
                },
                timeout=3,
            )
            assert ret["result"]
            assert ret["comment"] == "Data source test already up-to-date"
            assert ret["changes"] == {}

        with patch("requests.put") as rput:
            ret = grafana_datasource.present("test", "type", "newurl", profile=profile)
            rput.assert_called_once_with(
                "http://grafana/api/datasources/1",
                grafana_datasource._get_json_data("test", "type", "newurl"),
                headers={
                    "Authorization": "Bearer token",
                    "Accept": "application/json",
                },
                timeout=3,
            )
            assert ret["result"]
            assert ret["comment"] == "Data source test updated"
            assert ret["changes"] == {"old": {"url": "url"}, "new": {"url": "newurl"}}


def test_absent():
    with patch("requests.get", mock_json_response([])):
        with patch("requests.delete") as rdelete:
            ret = grafana_datasource.absent("test", profile=profile)
            assert rdelete.call_count == 0
            assert ret["result"]
            assert ret["comment"] == "Data source test already absent"

    with patch("requests.get", mock_json_response([{"name": "test", "id": 1}])):
        with patch("requests.delete") as rdelete:
            ret = grafana_datasource.absent("test", profile=profile)
            rdelete.assert_called_once_with(
                "http://grafana/api/datasources/1",
                headers={
                    "Authorization": "Bearer token",
                    "Accept": "application/json",
                },
                timeout=3,
            )
            assert ret["result"]
            assert ret["comment"] == "Data source test was deleted"
